@@ -3,7 +3,7 @@
  * Purpose:     MrEd main file, including a hodge-podge of global stuff
  * Author:      Matthew Flatt
  * Created:     1995
- * Copyright:   (c) 2004-2008 PLT Scheme Inc.
+ * Copyright:   (c) 2004-2009 PLT Scheme Inc.
  * Copyright:   (c) 1995-2000, Matthew Flatt
  */
 
@@ -29,7 +29,6 @@
 #include "wx_buttn.h"
 #include "wx_messg.h"
 #include "wx_timer.h"
-#include "wx_media.h"
 #include "wx_dialg.h"
 #include "wx_cmdlg.h"
 #include "wx_menu.h"
@@ -116,7 +115,6 @@ wxFrame *mred_real_main_frame;
 
 static Scheme_Thread *user_main_thread;
 
-extern void wxMediaIOCheckLSB(void);
 extern void wxMouseEventHandled(void);
 #ifdef wx_xt
 extern int wx_single_instance;
@@ -296,8 +294,6 @@ static int mark_eventspace_val(void *p)
   gcMARK_TYPED(MrEdFinalizedContext *, c->finalized);
 
   gcMARK_TYPED(wxChildList *, c->topLevelWindowList);
-  gcMARK_TYPED(wxStandardSnipClassList *, c->snipClassList);
-  gcMARK_TYPED(wxBufferDataClassList *, c->bufferDataClassList);
   gcMARK_TYPED(wxWindow *, c->modal_window);
   gcMARK_TYPED(MrEd_Saved_Modal *, c->modal_stack);
 
@@ -330,8 +326,6 @@ static int fixup_eventspace_val(void *p)
   gcFIXUP_TYPED(MrEdFinalizedContext *, c->finalized);
 
   gcFIXUP_TYPED(wxChildList *, c->topLevelWindowList);
-  gcFIXUP_TYPED(wxStandardSnipClassList *, c->snipClassList);
-  gcFIXUP_TYPED(wxBufferDataClassList *, c->bufferDataClassList);
   gcFIXUP_TYPED(wxWindow *, c->modal_window);
   gcFIXUP_TYPED(MrEd_Saved_Modal *, c->modal_stack);
 
@@ -523,22 +517,6 @@ void wxPopModalWindow(wxObject *w, wxWindow *win)
   }
 }
 
-wxStandardSnipClassList *wxGetTheSnipClassList()
-{
-  MrEdContext *c;
-  c = MrEdGetContext();
-
-  return c->snipClassList;
-}
-
-wxBufferDataClassList *wxGetTheBufferDataClassList()
-{
-  MrEdContext *c;
-  c = MrEdGetContext();
-
-  return c->bufferDataClassList;
-}
-
 int wxGetBusyState(void)
 {
   MrEdContext *c;
@@ -739,8 +717,6 @@ static MrEdContext *MakeContext(MrEdContext *c)
 
   if (!c) {
     wxChildList *tlwl;
-    wxStandardSnipClassList *scl;
-    wxBufferDataClassList *bdcl;
     MrEdFinalizedContext *fc;
 
     c = (MrEdContext *)scheme_malloc_tagged(sizeof(MrEdContext));
@@ -748,10 +724,6 @@ static MrEdContext *MakeContext(MrEdContext *c)
 
     tlwl = new WXGC_PTRS wxChildList();
     c->topLevelWindowList = tlwl;
-    scl = wxMakeTheSnipClassList();
-    c->snipClassList = scl;
-    bdcl = wxMakeTheBufferDataClassList();
-    c->bufferDataClassList = bdcl;
     fc = new WXGC_PTRS MrEdFinalizedContext;
     c->finalized = fc;
   }
@@ -1034,18 +1006,23 @@ static int check_eventspace_inactive(void *_c)
 void mred_wait_eventspace(void)
 {
   MrEdContext *c;
+  Scheme_Thread *thread;
   c = MrEdGetContext();
-  if (c && (c->handler_running == scheme_current_thread)) {
+  thread = scheme_get_current_thread();
+  if (c && (c->handler_running == thread)) {
     wxDispatchEventsUntilWaitable(check_eventspace_inactive, c, NULL);
   }
 }
 
 int mred_current_thread_is_handler(void *ctx)
 {
+  Scheme_Thread *thread;
+  thread = scheme_get_current_thread();
+
   if (!ctx)
     ctx = MrEdGetContext();
 
-  return (((MrEdContext *)ctx)->handler_running == scheme_current_thread);
+  return (((MrEdContext *)ctx)->handler_running == thread);
 }
 
 int mred_in_restricted_context()
@@ -1053,7 +1030,7 @@ int mred_in_restricted_context()
 #ifdef NEED_HET_PARAM
   /* see wxHiEventTrampoline for info on mred_het_key: */
   Scheme_Object *v;
-  if (!scheme_current_thread) 
+  if (!scheme_get_current_thread()) 
     return 1;
   
   if (mred_het_key)
@@ -1098,6 +1075,8 @@ static void DoTimer(wxTimer *timer)
 {
   int once;
   mz_jmp_buf *save, newbuf;
+  Scheme_Thread *thread;
+  thread = scheme_get_current_thread();
 
   if (timer->interval == -1)
     return;
@@ -1105,12 +1084,14 @@ static void DoTimer(wxTimer *timer)
   once = timer->one_shot;
   timer->one_shot = -1;
 
-  save = scheme_current_thread->error_buf;
-  scheme_current_thread->error_buf = &newbuf;
+  save = thread->error_buf;
+  thread->error_buf = &newbuf;
   if (!scheme_setjmp(newbuf))
     timer->Notify();
   scheme_clear_escape();
-  scheme_current_thread->error_buf = save;
+  thread = scheme_get_current_thread();
+  thread->error_buf = save;
+  thread = NULL;
 
   if (!once && (timer->one_shot == -1) && (timer->interval != -1)
       && !((MrEdContext *)timer->context)->killed)
@@ -1159,15 +1140,19 @@ static void GoAhead(MrEdContext *c)
   } else {
     GC_CAN_IGNORE MrEdEvent e;
     mz_jmp_buf *save, newbuf;
+    Scheme_Thread *thread;
+    thread = scheme_get_current_thread();
 
     memcpy(&e, &c->event, sizeof(MrEdEvent));
 
-    save = scheme_current_thread->error_buf;
-    scheme_current_thread->error_buf = &newbuf;
+    save = thread->error_buf;
+    thread->error_buf = &newbuf;
     if (!scheme_setjmp(newbuf))
       MrEdDispatchEvent(&e);
     scheme_clear_escape();
-    scheme_current_thread->error_buf = save;
+    thread = scheme_get_current_thread();
+    thread->error_buf = save;
+    thread = NULL;
   }
 }
 
@@ -1199,15 +1184,19 @@ static void DoTheEvent(MrEdContext *c)
   if (p != def_dispatch) {
     Scheme_Object *a[1];
     mz_jmp_buf *save, newbuf;
+    Scheme_Thread *thread;
+    thread = scheme_get_current_thread();
 
     a[0] = (Scheme_Object *)c;
 
-    save = scheme_current_thread->error_buf;
-    scheme_current_thread->error_buf = &newbuf;
+    save = thread->error_buf;
+    thread->error_buf = &newbuf;
     if (!scheme_setjmp(newbuf))
       scheme_apply_multi(p, 1, a);
     scheme_clear_escape();
-    scheme_current_thread->error_buf = save;
+    thread = scheme_get_current_thread();
+    thread->error_buf = save;
+    thread = NULL;
 
 #if 0
     if (c->ready_to_go)
@@ -1334,10 +1323,12 @@ static Scheme_Object *MrEdDoNextEvent(MrEdContext *c, wxDispatch_Check_Fun alt,
 void wxDoNextEvent()
 {
   MrEdContext *c;
+  Scheme_Thread *thread;
   c = MrEdGetContext();
+  thread = scheme_get_current_thread();
 
   if (!c->ready_to_go)
-    if (c->handler_running == scheme_current_thread)
+    if (c->handler_running == thread)
       MrEdDoNextEvent(c, NULL, NULL, NULL);
 }
 
@@ -1362,10 +1353,12 @@ int MrEdEventReady(MrEdContext *c)
 int wxEventReady()
 {
   MrEdContext *c;
+  Scheme_Thread *thread;
   c = MrEdGetContext();
+  thread = scheme_get_current_thread();
 
   return (!c->ready_to_go
-	  && (c->handler_running == scheme_current_thread)
+	  && (c->handler_running == thread)
 	  && MrEdEventReady(c));
 }
 
@@ -1389,7 +1382,7 @@ static void WaitForAnEvent_OrDie(MrEdContext *c)
       c->waiting_for_nested = 0;
 
       scheme_thread_block(0);
-      scheme_current_thread->ran_some = 1;
+      scheme_set_current_thread_ran_some();
 
       /* Go back to sleep: */
       c->ready = 1;
@@ -1435,7 +1428,7 @@ static Scheme_Object *handle_events(void *cx, int, Scheme_Object **)
   fprintf(stderr, "new thread\n");
 #endif
 
-  this_thread = scheme_current_thread;
+  this_thread = scheme_get_current_thread();
   if (!this_thread->name) {
     Scheme_Object *tn;
     tn = scheme_intern_symbol("handler");
@@ -1447,7 +1440,7 @@ static Scheme_Object *handle_events(void *cx, int, Scheme_Object **)
   c->suspended = 0;
   c->ready = 0;
 
-  scheme_current_thread->error_buf = &newbuf;
+  this_thread->error_buf = &newbuf;
   if (!scheme_setjmp(newbuf)) {
     if (!TheMrEdApp->initialized)
       TheMrEdApp->RealInit();
@@ -1512,7 +1505,7 @@ static int try_q_callback(Scheme_Object *do_it, int hi)
       return 1;
 
     if (SCHEME_FALSEP(do_it))
-      scheme_current_thread->ran_some = 1;
+      scheme_set_current_thread_ran_some();
 
     if (c == mred_main_context)
       check_q_callbacks(hi, MrEdSameContext, c, 0);
@@ -1543,7 +1536,7 @@ static int try_dispatch(Scheme_Object *do_it)
     if (!do_it)
       return 1;
     if (SCHEME_FALSEP(do_it))
-      scheme_current_thread->ran_some = 1;
+      scheme_set_current_thread_ran_some();
 
     c = (MrEdContext *)timer->context;
 
@@ -1579,7 +1572,7 @@ static int try_dispatch(Scheme_Object *do_it)
       return 1;
 
     if (SCHEME_FALSEP(do_it))
-      scheme_current_thread->ran_some = 1;
+      scheme_set_current_thread_ran_some();
 
     if (c) {
       memcpy(&c->event, &e, sizeof(MrEdEvent));
@@ -1617,27 +1610,11 @@ static int check_initialized(Scheme_Object *)
 
 # define KEEP_GOING wxTheApp->keep_going
 
-#if WINDOW_STDIO
-static Scheme_Custodian *main_custodian;
-#endif
-
 void wxDoEvents()
 {
   /* When we get here, we are in the main dispatcher thread */
   if (!TheMrEdApp->initialized) {
     MrEdContext *c;
-#if WINDOW_STDIO
-    Scheme_Custodian *m, *oldm = NULL;
-    Scheme_Config *config = NULL;
-    if (!wx_in_terminal) {
-      config = scheme_current_config();
-      oldm = (Scheme_Custodian *)scheme_get_param(config, MZCONFIG_CUSTODIAN);
-      m = scheme_make_custodian(oldm);
-      scheme_set_param(config, MZCONFIG_CUSTODIAN, (Scheme_Object *)m);
-      wxREGGLOB(main_custodian);
-      main_custodian = m;
-    }
-#endif
 
     /* Create the user's main thread: */
 
@@ -1660,25 +1637,23 @@ void wxDoEvents()
       user_main_thread->name = cp;
     }
 
-#if WINDOW_STDIO
-    if (!wx_in_terminal)
-      scheme_set_param(config, MZCONFIG_CUSTODIAN, (Scheme_Object *)oldm);
-#endif
-
     /* Block until the user's main thread is initialized: */
     scheme_block_until(CAST_BLKCHK check_initialized, NULL, NULL, 0.0);
   }
 
   if (!try_dispatch(scheme_true)) {
     do {
-      scheme_current_thread->block_descriptor = -1;
-      scheme_current_thread->blocker = NULL;
-      scheme_current_thread->block_check = CAST_BLKCHK try_dispatch;
-      scheme_current_thread->block_needs_wakeup = CAST_WU wakeup_on_dispatch;
+      Scheme_Thread *thread;
+      thread = scheme_get_current_thread();
+      thread->block_descriptor = -1;
+      thread->blocker = NULL;
+      thread->block_check = CAST_BLKCHK try_dispatch;
+      thread->block_needs_wakeup = CAST_WU wakeup_on_dispatch;
 
       scheme_thread_block(0);
 
-      scheme_current_thread->block_descriptor = 0;
+      thread = scheme_get_current_thread();
+      thread->block_descriptor = 0;
       /* Sets ran_some if it succeeds: */
       if (try_dispatch(scheme_false))
 	break;
@@ -1690,14 +1665,16 @@ Scheme_Object *wxDispatchEventsUntilWaitable(wxDispatch_Check_Fun f, void *data,
 {
   MrEdContext *c;
   Scheme_Object *result = scheme_void;
+  Scheme_Thread *thread;
 
   c = MrEdGetContext();
 #ifdef wx_mac
   wxMouseEventHandled();
 #endif
 
+  thread = scheme_get_current_thread();
   if (c->ready_to_go
-      || (c->handler_running != scheme_current_thread)) {
+      || (c->handler_running != thread)) {
     /* This is not the handler thread or an event still hasn't been
        dispatched. Wait. */
     if (w) {
@@ -1766,7 +1743,7 @@ static void MrEdSleep(float secs, void *fds)
   }
 
 #ifdef wx_msw
-  MrEdMSWSleep(secs, fds);
+  MrEdMSWSleep(secs, fds, mzsleep);
 #else
 # ifdef wx_mac
   MrEdMacSleep(secs, fds, mzsleep);
@@ -1956,13 +1933,16 @@ static void remove_q_callback(Q_Callback_Set *cs, Q_Callback *cb)
 static void call_one_callback(Q_Callback * volatile  cb)
 {
   mz_jmp_buf *save, newbuf;
+  Scheme_Thread *thread;
+  thread = scheme_get_current_thread();
 
-  save = scheme_current_thread->error_buf;
-  scheme_current_thread->error_buf = &newbuf;
+  save = thread->error_buf;
+  thread->error_buf = &newbuf;
   if (!scheme_setjmp(newbuf))
     scheme_apply_multi(cb->callback, 0, NULL);
   scheme_clear_escape();
-  scheme_current_thread->error_buf = save;
+  thread = scheme_get_current_thread();
+  thread->error_buf = save;
 }
 
 static MrEdContext *check_q_callbacks(int hi, int (*test)(MrEdContext *, MrEdContext *),
@@ -2042,7 +2022,7 @@ static void MrEdQueueWindowCallback(wxWindow *wx_window, Scheme_Closed_Prim *scp
   Q_Callback *cb;
   Scheme_Object *p;
 
-  if (!scheme_current_thread) {
+  if (!scheme_get_current_thread()) {
     /* Scheme hasn't started yet, so call directly.
        We might get here for an update to the stdio
        window, for example. */
@@ -2226,236 +2206,52 @@ void MrEdQueueInEventspace(void *context, Scheme_Object *thunk)
 /*                        Redirected Standard I/O                           */
 /****************************************************************************/
 
-#if REDIRECT_STDIO || WINDOW_STDIO || WCONSOLE_STDIO
+#if REDIRECT_STDIO || WCONSOLE_STDIO
 static void MrEdSchemeMessages(char *, ...);
 static Scheme_Object *stdin_pipe;
 #endif
 
-#if WINDOW_STDIO
-
-static int have_stdio = 0;
-static int stdio_kills_prog = 0;
-static Bool RecordInput(void *media, wxEvent *event, void *data);
-static Bool SendBreak(void *media, wxEvent *event, void *data);
-static void break_console_reading_threads();
-static char utf8_leftover[8];
-static int utf8_leftover_count;
-
-class IOMediaEdit : public wxMediaEdit
-{
-public:
-    Bool CanInsert(long start, long);
-    Bool CanDelete(long start, long);
-};
-
-class IOFrame : public wxFrame
-{
-public:
-  wxMediaCanvas *display;
-  wxMediaEdit *media;
-  wxMenu *fileMenu;
-  Bool hidden, beginEditSeq;
-  int endpos;
-
-  IOFrame();
-  void OnSize(int x, int y);
-  Bool OnClose(void);
-  void OnMenuCommand(long id);
-  Bool PreOnChar(wxWindow *, wxKeyEvent *e);
-  Bool PreOnEvent(wxWindow *, wxMouseEvent *e);
-  void CloseIsQuit(void);
-};
-
-IOFrame::IOFrame()
- : wxFrame(NULL, "Standard Output", -1, -1, 600, 400, 0, "stdout")
-{
-  wxKeymap *km;
-  wxStyle *style;
-  wxStyleList *sl;
-  wxStyleDelta *sd;
-  wxMenuBar *mb;
-  wxMenu *m;
-
-  display = new WXGC_PTRS wxMediaCanvas(this);
-
-  media = new WXGC_PTRS IOMediaEdit();
-  display->SetMedia(media);
-  endpos = 0;
-  hidden = FALSE;
-
-  /* Map copy keys: */
-  km = media->GetKeymap();
-  media->AddBufferFunctions(km);
-  media->AddEditorFunctions(km);
-  km->AddFunction("send-break", SendBreak, NULL);
-# ifdef wx_msw
-  km->MapFunction("c:c", "copy-clipboard");
-  km->MapFunction("c:x", "copy-clipboard");
-  km->MapFunction("c:v", "paste-clipboard");
-# else
-  km->MapFunction("d:c", "copy-clipboard");
-  km->MapFunction("d:x", "copy-clipboard");
-  km->MapFunction("d:v", "paste-clipboard");
-  km->MapFunction("d:.", "send-break");
-# endif
-  km->MapFunction("return", "record-input");
-  km->AddFunction("record-input", RecordInput, NULL);
-
-  /* Fixed-width font: */
-  sl = media->GetStyleList();
-  style = sl->FindNamedStyle("Standard");
-  sd = new WXGC_PTRS wxStyleDelta(wxCHANGE_FAMILY, wxMODERN);
-  style->SetDelta(sd);
-
-#ifdef wx_mac
-  OnSize(600, 400);
-#endif
-
-#ifdef wx_mac
-# define CLOSE_MENU_ITEM "Close\tCmd+W"
-#else
-# define CLOSE_MENU_ITEM "Close"
-#endif
-
-  mb = new WXGC_PTRS wxMenuBar();
-  SetMenuBar(mb);
-  fileMenu = new WXGC_PTRS wxMenu();
-  fileMenu->Append(77, CLOSE_MENU_ITEM);
-  m = new WXGC_PTRS wxMenu();
-  m->Append(79, "&Copy\tCmd+C");
-  m->Append(81, "&Paste\tCmd+V");
-  m->AppendSeparator();
-  m->Append(83, "&Break\tCmd+.");
-  mb->Append(fileMenu, "File");
-  mb->Append(m, "Edit");
-
-  have_stdio = 1;
-  Show(TRUE);
-
-  beginEditSeq = 0;
-}
-
-void IOFrame::OnSize(int x, int y)
-{
-  GetClientSize(&x, &y);
-  if (display)
-    display->SetSize(0, 0, x, y);
-  if (media && (x > 30))
-    media->SetMaxWidth((float)(x - 30));
-}
-
-Bool IOFrame::OnClose(void)
-{
-  hidden = TRUE;
-  if (stdio_kills_prog) {
-    if (scheme_exit)
-      scheme_exit(exit_val);
-#ifdef wx_msw
-    mred_clean_up_gdi_objects();
-#endif
-    scheme_immediate_exit(exit_val);
-  } else {
-    break_console_reading_threads();
-    have_stdio = 0;
-  }
-  return TRUE;
-}
-
-void IOFrame::OnMenuCommand(long id)
-{
-  if (id == 79)
-    media->Copy();
-  else if (id == 81)
-    media->Paste();
-  else if (id == 83)
-    scheme_break_main_thread();
-  else if (id == 77)
-    if (OnClose())
-      Show(FALSE);
-}
-
-Bool IOFrame::PreOnChar(wxWindow *, wxKeyEvent *e)
-{
-  PreOnEvent(NULL, NULL);
-
-#if defined(wx_mac) && WINDOW_STDIO
-  if (e->metaDown && e->KeyCode() == (stdio_kills_prog ? 'q' : 'w')) {
-    OnMenuCommand(77);
-    return TRUE;
-  }
-#endif
-
-  return FALSE;
-}
-
-Bool IOFrame::PreOnEvent(wxWindow *, wxMouseEvent *e)
-{
-  if (beginEditSeq) {
-    beginEditSeq = 0;
-    media->EndEditSequence();
-  }
-
-  return FALSE;
-}
-
-void IOFrame::CloseIsQuit(void)
-{
-#ifdef wx_mac
-# define QUIT_MENU_ITEM "Quit\tCmd+Q"
-#else
-# define QUIT_MENU_ITEM "E&xit"
-#endif
-  fileMenu->Delete(77);
-  fileMenu->Append(77, QUIT_MENU_ITEM);
-
-  media->Insert("\n[Exited]", media->LastPosition());
-  if (beginEditSeq) {
-    beginEditSeq = 0;
-    media->EndEditSequence();
-  }
-  media->Lock(1);
-}
-
-static IOFrame *ioFrame = NULL;
-
-Bool IOMediaEdit::CanInsert(long start, long)
-{
-    return (start >= ioFrame->endpos);
-}
-
-Bool IOMediaEdit::CanDelete(long start, long)
-{
-    return (start >= ioFrame->endpos);
-}
-
-static Bool RecordInput(void *m, wxEvent *event, void *data)
-{
-  char *s;
-  long len, start;
-  wxMediaEdit *media = ioFrame->media;
-
-  media->Insert("\n");
-  start = media->GetStartPosition();
-  len = start - ioFrame->endpos;
-  s = media->GetTextUTF8(ioFrame->endpos, start);
-  ioFrame->endpos = start;
-
-  scheme_write_byte_string(s, len, stdin_pipe);
-
-  return TRUE;
-}
-
-static Bool SendBreak(void *m, wxEvent *event, void *data)
-{
-  scheme_break_main_thread();
-  return TRUE;
-}
-
-#else  /* !WINDOW_STDIO */
-
 #if WCONSOLE_STDIO
 
 static HANDLE console_out;
+static HANDLE console_in;
+static Scheme_Object *console_inport;
+static HWND console_hwnd;
+static int has_stdio, stdio_kills_prog;
+static HANDLE waiting_sema;
+
+typedef HWND (WINAPI* gcw_proc)();
+
+static BOOL WINAPI ConsoleHandler(DWORD op)
+{
+  if (stdio_kills_prog) {
+    ReleaseSemaphore(waiting_sema, 1, NULL);
+  } else {
+    scheme_break_main_thread();
+    scheme_signal_received();
+  }
+  return TRUE;
+}
+
+static void WaitOnConsole()
+{
+  DWORD wrote;
+
+  stdio_kills_prog = 1;
+  if (console_hwnd) {
+    AppendMenu(GetSystemMenu(console_hwnd, FALSE), 
+	       MF_STRING,
+	       SC_CLOSE,
+	       "Close");
+    /* Un-gray the close box: */
+    RedrawWindow(console_hwnd, NULL, NULL, 
+		 RDW_FRAME | RDW_INVALIDATE | RDW_UPDATENOW);
+  }
+
+  WriteConsole(console_out, "\n[Exited. Close box or Ctrl-C closes the console.]\n", 51, &wrote, NULL);
+
+  WaitForSingleObject(waiting_sema, INFINITE);
+}
 
 #else  /* !WCONSOLE_STDIO */
 
@@ -2467,70 +2263,49 @@ static FILE *mrerr = NULL;
 
 #endif /* WCONSOLE_STDIO */
 
-#endif /* WINDOW_STDIO */
-
-#if REDIRECT_STDIO || WINDOW_STDIO || WCONSOLE_STDIO
+#if REDIRECT_STDIO || WCONSOLE_STDIO
 static void MrEdSchemeMessages(char *msg, ...)
 {
   GC_CAN_IGNORE va_list args;
-#if WINDOW_STDIO
-  char *arg_s = NULL;
-  long arg_d = 0, arg_l = 0;
-# define VSP_BUFFER_SIZE 4096
-  char arg_buffer[VSP_BUFFER_SIZE];
-#endif
   
   scheme_start_atomic();
 
   HIDE_FROM_XFORM(va_start(args, msg));
 
-#if WINDOW_STDIO
-  if (!wx_in_terminal) {
-    static int opening = 0;
-
-    if (opening) {
-      HIDE_FROM_XFORM(va_end(args));
-      return;
-    }
-
-    /* Need to extract arguments before a potential GC */
-    if (!msg) {
-      arg_s = HIDE_FROM_XFORM(va_arg(args, char*));
-      arg_d = HIDE_FROM_XFORM(va_arg(args, long));
-      arg_l = HIDE_FROM_XFORM(va_arg(args, long));
-    } else {
-# ifdef MPW_CPLUS
-      /* FIXME: No vsnprintf in MPW. */
-#  define vsnprintf(x, y, z, w) vsprintf(x, z, w)
-# endif
-      MSC_IZE(vsnprintf)(arg_buffer, VSP_BUFFER_SIZE, msg, args);
-    }
-
-    opening = 1;
-    if (!ioFrame) {
-      wxREGGLOB(ioFrame);
-      if (mred_only_context)
-	ioFrame = new WXGC_PTRS IOFrame;
-      else {
-	/* Set eventspace ... */
-	mred_only_context = mred_main_context;
-	only_context_just_once = 1;
-	ioFrame = new WXGC_PTRS IOFrame;
-	mred_only_context = NULL;
-      }
-    }
-    opening = 0;
-    if (ioFrame->hidden) {
-      ioFrame->hidden = FALSE;
-      have_stdio = 1;
-      ioFrame->Show(TRUE);
-    }
-  }
-#endif
 #if WCONSOLE_STDIO
   if (!console_out) {
     AllocConsole();
     console_out = GetStdHandle(STD_OUTPUT_HANDLE);
+
+    if (!wx_in_terminal) {
+      console_in = GetStdHandle(STD_INPUT_HANDLE);
+      has_stdio = 1;
+      waiting_sema = CreateSemaphore(NULL, 0, 1, NULL);
+      SetConsoleCtrlHandler(ConsoleHandler, TRUE);
+
+      wxREGGLOB(console_inport);
+      console_inport = scheme_make_fd_input_port((int)console_in, scheme_intern_symbol("stdin"), 0, 0);
+
+      {
+	HMODULE hm;
+	gcw_proc gcw;
+
+	hm = LoadLibrary("kernel32.dll");
+	if (hm)
+	  gcw = (gcw_proc)GetProcAddress(hm, "GetConsoleWindow");
+	else
+	  gcw = NULL;
+    
+	if (gcw)
+	  console_hwnd = gcw();
+      }
+
+      if (console_hwnd) {
+	EnableMenuItem(GetSystemMenu(console_hwnd, FALSE), SC_CLOSE,
+		       MF_BYCOMMAND | MF_GRAYED);
+	RemoveMenu(GetSystemMenu(console_hwnd, FALSE), SC_CLOSE, MF_BYCOMMAND);
+      }
+    }
   }
 #endif
 #if REDIRECT_STDIO
@@ -2543,78 +2318,27 @@ static void MrEdSchemeMessages(char *msg, ...)
   }
 #endif
 
-#if WINDOW_STDIO
-  if (wx_in_terminal) {
-    vfprintf(stderr, msg, args);
-    fflush(stderr);
-  } else if (!msg) {
-    char *s;
-    wxchar *us;
-    long d, l, ulen, ipos;
-
-    s = arg_s;
-    d = arg_d;
-    l = arg_l;
-
-    if (!ioFrame->beginEditSeq) {
-      ioFrame->media->BeginEditSequence();
-      ioFrame->beginEditSeq = 1;
-    }
-    
-    if (utf8_leftover_count) {
-      char *naya;
-      naya = new WXGC_ATOMIC char[l + utf8_leftover_count];
-      memcpy(naya, utf8_leftover, utf8_leftover_count);
-      memcpy(naya + utf8_leftover_count, s + d, l);
-      s = naya;
-      d = 0;
-      l += utf8_leftover_count;
-    }
-
-    ulen = scheme_utf8_decode_as_prefix((unsigned char *)s, d, l,
-					NULL, 0, -1,
-					&ipos, 0, '?');
-    utf8_leftover_count = (l - (ipos - d));
-    memcpy(utf8_leftover, s + ipos, utf8_leftover_count);
-    
-    us = (wxchar *)scheme_malloc_atomic(sizeof(wxchar) * ulen);
-    scheme_utf8_decode_as_prefix((unsigned char *)s, d, l,
-				 us, 0, -1,
-				 &ipos, 0, '?');
-    ioFrame->media->Insert(ulen, us, ioFrame->endpos);
-    ioFrame->endpos += ulen;
-
-    if (ulen != 1 || s[0] == '\n') {
-      ioFrame->media->EndEditSequence();
-      ioFrame->beginEditSeq = 0;
-    }
-  } else {
-    ioFrame->media->Insert((char *)arg_buffer, ioFrame->endpos);
-    ioFrame->endpos += strlen(arg_buffer);
-    if (ioFrame->beginEditSeq) {
-      ioFrame->media->EndEditSequence();
-      ioFrame->beginEditSeq = 0;
-    }
-  }
-#endif
 #if WCONSOLE_STDIO
   if (!msg) {
     char *s;
-    long l;
-	DWORD wrote;
+    long l, d;
+    DWORD wrote;
 
     s = va_arg(args, char*);
+    d = va_arg(args, long);
     l = va_arg(args, long);
 
-    WriteConsole(console_out, s, l, &wrote, NULL);
+    WriteConsole(console_out, s XFORM_OK_PLUS d, l, &wrote, NULL);
   } else {
-    char buffer[2048];
+    char *buffer;
     DWORD wrote;
+    buffer = (char *)malloc(5 * strlen(msg));
     vsprintf(buffer, msg, args);
     WriteConsole(console_out, buffer, strlen(buffer), &wrote, NULL);
+    free(buffer);
   }
 #endif
-#if !WINDOW_STDIO && !WCONSOLE_STDIO
+#if !WCONSOLE_STDIO
   vfprintf(mrerr, msg, args);
   fflush(mrerr);
 #endif
@@ -2631,31 +2355,36 @@ static void MrEdSchemeMessagesOutput(char *s, long l)
 }
 #endif
 
-#if REDIRECT_STDIO || WINDOW_STDIO || WCONSOLE_STDIO
+#if REDIRECT_STDIO || WCONSOLE_STDIO
 
 static Scheme_Object *console_reading;
 
 static void add_console_reading()
 {
+  Scheme_Thread *thread;
+  thread = scheme_get_current_thread();
+
   if (!console_reading) {
     wxREGGLOB(console_reading);
     console_reading = scheme_null;
   }
 
-  console_reading = scheme_make_pair((Scheme_Object *)scheme_current_thread,
+  console_reading = scheme_make_pair((Scheme_Object *)thread,
 				     console_reading);
 }
 
 static void remove_console_reading()
 {
   Scheme_Object *p, *prev = NULL;
+  Scheme_Thread *thread;
+  thread = scheme_get_current_thread();
 
   if (!console_reading)
     return;
 
   p = console_reading;
   while (SCHEME_PAIRP(p)) {
-    if (SAME_OBJ(SCHEME_CAR(p), (Scheme_Object *)scheme_current_thread)) {
+    if (SAME_OBJ(SCHEME_CAR(p), (Scheme_Object *)thread)) {
       if (prev)
 	SCHEME_CDR(prev) = SCHEME_CDR(p);
       else
@@ -2687,22 +2416,64 @@ static long mrconsole_get_string(Scheme_Input_Port *ip,
   Scheme_Object *pipe = (Scheme_Object *)ip->port_data;
   MrEdSchemeMessages("");
 
+#if WCONSOLE_STDIO
+  pipe = console_inport;
+#endif
+
   add_console_reading();
-  result = scheme_get_byte_string("console get-string", pipe, buffer, offset, size, nonblock ? 2 : 0, 0, 0);
+  result = scheme_get_byte_string_unless("console get-string", pipe, 
+					 buffer, offset, size, 
+					 nonblock, 0, NULL,
+					 unless);
   remove_console_reading();
   return result;
 }
 
+static Scheme_Object *mrconsole_progress_evt(Scheme_Input_Port *ip)
+{
+  Scheme_Object *pipe = (Scheme_Object *)ip->port_data;
+  MrEdSchemeMessages("");
+
+#if WCONSOLE_STDIO
+  pipe = console_inport;
+#endif
+
+  return scheme_progress_evt(pipe);
+}
+
+static int mrconsole_peeked_read(Scheme_Input_Port *ip,
+					    long amount,
+					    Scheme_Object *unless,
+					    Scheme_Object *target_ch)
+{
+  Scheme_Object *pipe = (Scheme_Object *)ip->port_data;
+  MrEdSchemeMessages("");
+
+#if WCONSOLE_STDIO
+  pipe = console_inport;
+#endif
+
+  return scheme_peeked_read(pipe, amount, unless, target_ch);
+}
+
 static int mrconsole_char_ready(Scheme_Input_Port *ip)
 {
   Scheme_Object *pipe = (Scheme_Object *)ip->port_data;
   MrEdSchemeMessages("");
+
+#if WCONSOLE_STDIO
+  pipe = console_inport;
+#endif
+
   return scheme_char_ready(pipe);
 }
 
 static void mrconsole_close(Scheme_Input_Port *ip)
 {
   Scheme_Object *pipe = (Scheme_Object *)ip->port_data;
+#if WCONSOLE_STDIO
+  pipe = console_inport;
+#endif
   scheme_close_input_port(pipe);
 }
 
@@ -2720,8 +2491,8 @@ static Scheme_Object *MrEdMakeStdIn(void)
 			      scheme_intern_symbol("mred-console"),
 			      CAST_GS mrconsole_get_string,
 			      NULL,
-			      scheme_progress_evt_via_get,
-			      scheme_peeked_read_via_get,
+			      mrconsole_progress_evt,
+			      mrconsole_peeked_read,
 			      CAST_IREADY mrconsole_char_ready,
 			      CAST_ICLOSE mrconsole_close,
 			      NULL,
@@ -2733,7 +2504,7 @@ static Scheme_Object *MrEdMakeStdIn(void)
 static long stdout_write(Scheme_Output_Port*, const char *s, long d, long l, 
 			 int rarely_block, int enable_break)
 {
-#if WINDOW_STDIO || WCONSOLE_STDIO
+#if WCONSOLE_STDIO
   if (l)
     MrEdSchemeMessages(NULL, s, d, l);
 #else
@@ -2764,7 +2535,7 @@ static Scheme_Object *MrEdMakeStdOut(void)
 static long stderr_write(Scheme_Output_Port*, const char *s, long d, long l, 
 			 int rarely_block, int enable_break)
 {
-#if WINDOW_STDIO || WCONSOLE_STDIO
+#if WCONSOLE_STDIO
   if (l)
     MrEdSchemeMessages(NULL, s, d, l);
 #else
@@ -2813,7 +2584,7 @@ extern "C" GC_word GC_fo_entries;
 
 Scheme_Object *OBJDump(int, Scheme_Object *[])
 {
-# if REDIRECT_STDIO || WINDOW_STDIO || WCONSOLE
+# if REDIRECT_STDIO || WCONSOLE_STDIO
 # define PRINT_IT MrEdSchemeMessages
 # else
 # define PRINT_IT scheme_console_printf
@@ -3114,20 +2885,24 @@ extern "C" {
 
 static void MrEdOutOfMemory(void)
 {
+  /* Hopefully we have enough memory for a message dialog under
+     Windows and Mac OS X: */
 #ifdef wx_mac
   Alert(101, NULL);
-  ExitToShell();
-#else
-#ifdef wx_x
-  printf("mred: out of memory\n");
 #endif
-  _exit(-1);
+#ifdef wx_msw
+  wxNoMoreCallbacks();
+  MessageBox(NULL, 
+             "PLT Scheme virtual machine is out of memory. Aborting.",
+             "Out of Memory",
+             MB_OK);
 #endif
+  /* For X, mzscheme already writes to stderr (and maybe syslog). */
 }
 
 void *wxOutOfMemory()
 {
-  MrEdOutOfMemory();
+  scheme_out_of_memory_abort();
   return NULL;
 }
 
@@ -3162,19 +2937,6 @@ void *wxMallocAtomicIfPossible(size_t s)
   return v;
 }
 
-static const char *CallSchemeExpand(const char *filename, const char *who, int to_write)
-{
-  char *s;
-
-  s = scheme_expand_filename((char *)filename, strlen(filename),
-			     who, 0,
-			     (to_write
-			      ? SCHEME_GUARD_FILE_WRITE
-			      : SCHEME_GUARD_FILE_READ));
-
-  return s ? s : filename;
-}
-
 #if !defined(USE_SENORA_GC) && !defined(MZ_PRECISE_GC)
 static void MrEdIgnoreWarnings(char *, GC_word)
 {
@@ -3191,12 +2953,8 @@ static Scheme_Env *setup_basic_env()
   wxREGGLOB(global_env);
   global_env = scheme_basic_env();
 
-  scheme_no_dumps("the graphics library is running");
-
   scheme_set_banner(BANNER);
 
-  wxmeExpandFilename = CallSchemeExpand;
-
 #ifdef DANGER_ALARM
   {
     DangerThreadTimer *t = new WXGC_PTRS DangerThreadTimer();
@@ -3236,7 +2994,11 @@ static Scheme_Env *setup_basic_env()
 
   /* This handler_running pointer gets reset later. Do
      we really need to set it now? */
-  mred_main_context->handler_running = scheme_current_thread;
+  {
+    Scheme_Thread *thread;
+    thread = scheme_get_current_thread();
+    mred_main_context->handler_running = thread;
+  }
 
   mzsleep = scheme_sleep;
   scheme_sleep = CAST_SLEEP MrEdSleep;
@@ -3249,6 +3011,20 @@ static Scheme_Env *setup_basic_env()
   return global_env;
 }
 
+#if WCONSOLE_STDIO
+static void MrEdExit(int v)
+{
+  if (has_stdio) {
+    WaitOnConsole();
+  }
+
+#ifdef wx_msw
+  mred_clean_up_gdi_objects();
+#endif
+  scheme_immediate_exit(v);
+}
+#endif
+
 wxFrame *MrEdApp::OnInit(void)
 {
   MrEdContext *mmc;
@@ -3274,7 +3050,7 @@ wxFrame *MrEdApp::OnInit(void)
   new WXGC_PTRS Regex("a", 0);
 #endif
 
-#if REDIRECT_STDIO || WINDOW_STDIO || WCONSOLE_STDIO
+#if REDIRECT_STDIO || WCONSOLE_STDIO
   if (!wx_in_terminal) {
     scheme_make_stdin = CAST_MK MrEdMakeStdIn;
     scheme_make_stdout = CAST_MK MrEdMakeStdOut;
@@ -3285,12 +3061,7 @@ wxFrame *MrEdApp::OnInit(void)
 #if !defined(USE_SENORA_GC) && !defined(MZ_PRECISE_GC)
   GC_set_warn_proc(CAST_IGNORE MrEdIgnoreWarnings);
 #endif
-#if 0
-  /* Used to be set for the sake of Mac OS Classic. Now,
-     setting GC_out_of_memory for 3m means that it's ok
-     to fail when a limit is reached. We don't want that. */
-  GC_out_of_memory = (OOM_ptr)MrEdOutOfMemory;
-#endif
+  scheme_set_report_out_of_memory(MrEdOutOfMemory);
 
 #ifdef SGC_STD_DEBUGGING
   scheme_external_dump_info = dump_cpp_info;
@@ -3300,7 +3071,7 @@ wxFrame *MrEdApp::OnInit(void)
 # endif
 #endif
 
-#if REDIRECT_STDIO || WINDOW_STDIO || WCONSOLE_STDIO
+#if REDIRECT_STDIO || WCONSOLE_STDIO
   scheme_console_printf = CAST_PRINTF MrEdSchemeMessages;
   if (!wx_in_terminal) {
     scheme_console_output = CAST_OUTPUT MrEdSchemeMessagesOutput;
@@ -3311,8 +3082,6 @@ wxFrame *MrEdApp::OnInit(void)
   mred_event_dispatch_param = scheme_new_param();
   mred_ps_setup_param = scheme_new_param();
 
-  wxInitSnips(); /* and snip classes */
-
   mred_eventspace_type = scheme_make_type("<eventspace>");
   mred_nested_wait_type = scheme_make_type("<eventspace-nested-wait>");
   mred_eventspace_hop_type = scheme_make_type("<internal:eventspace-hop>");
@@ -3352,16 +3121,6 @@ wxFrame *MrEdApp::OnInit(void)
     cl = new WXGC_PTRS wxChildList();
     mmc->topLevelWindowList = cl;
   }
-  {
-    wxStandardSnipClassList *scl;
-    scl = wxMakeTheSnipClassList();
-    mmc->snipClassList = scl;
-  }
-  {
-    wxBufferDataClassList *dcl;
-    dcl = wxMakeTheBufferDataClassList();
-    mmc->bufferDataClassList = dcl;
-  }
   {
     MrEdFinalizedContext *fc;
     fc = new WXGC_PTRS MrEdFinalizedContext;
@@ -3381,7 +3140,7 @@ wxFrame *MrEdApp::OnInit(void)
   TheMrEdApp->wx_frame = mred_real_main_frame;
 #endif
 
-  wxInitMedia();
+  wxInitClipboard();
 
   wxscheme_early_gl_init();
 
@@ -3391,45 +3150,29 @@ wxFrame *MrEdApp::OnInit(void)
 # else
   MZ_SIGSET(SIGINT, user_break_hit);
 # endif
+#endif
+
+#ifdef wx_mac
+# ifdef OS_X
+  /* Hack to make sure it's referenced, so that xform doesn't throw it away. */
+  wx_in_terminal = wx_in_terminal;
+# endif
 #endif
 
   mred_run_from_cmd_line(argc, argv, setup_basic_env);
 
-#if WINDOW_STDIO
+#if WCONSOLE_STDIO
   if (!wx_in_terminal) {
     /* The only reason we get here is that a command-line error or
        -h occured. In either case, stick around for the sake of the
        console. */
-    setup_basic_env();
-    TheMrEdApp->initialized = 1;
-    stdio_kills_prog = 1;
-    if (ioFrame)
-      ioFrame->CloseIsQuit();
-    wxTheApp->MainLoop();
+    MrEdExit(1);
   }
 #endif
 
   return NULL;
 }
 
-#if WINDOW_STDIO
-static void MrEdExit(int v)
-{
-  if (have_stdio) {
-    stdio_kills_prog = 1;
-    if (ioFrame)
-      ioFrame->CloseIsQuit();
-    scheme_close_managed(main_custodian);
-    return;
-  }
-
-#ifdef wx_msw
-  mred_clean_up_gdi_objects();
-#endif
-  scheme_immediate_exit(v);
-}
-#endif
-
 static void on_main_killed(Scheme_Thread *p)
 {
   on_handler_killed(p);
@@ -3446,12 +3189,13 @@ static void on_main_killed(Scheme_Thread *p)
 
 void MrEdApp::RealInit(void)
 {
+  Scheme_Thread *thread;
+  thread = scheme_get_current_thread();
+
   initialized = 1;
 
-  wxMediaIOCheckLSB(/* scheme_console_printf */);
-
-  scheme_current_thread->on_kill = CAST_TOK on_main_killed;
-#if WINDOW_STDIO
+  thread->on_kill = CAST_TOK on_main_killed;
+#if WCONSOLE_STDIO
   if (!wx_in_terminal)
     scheme_exit = CAST_EXIT MrEdExit;
 #endif
@@ -3465,7 +3209,7 @@ void MrEdApp::RealInit(void)
   if (!exit_val)
     exit_val = mred_finish_cmd_line_run();
 
-  scheme_kill_thread(scheme_current_thread);
+  scheme_kill_thread(thread);
 }
 
 #ifdef wx_mac
@@ -3878,6 +3622,8 @@ static void wxDo(Scheme_Object *proc, int argc, Scheme_Object **argv)
 {
   mz_jmp_buf * volatile save, newbuf;
   volatile int block_descriptor;
+  Scheme_Thread *thread;
+  thread = scheme_get_current_thread();
 
   if (!proc) {
     /* Oops --- too early. */
@@ -3887,13 +3633,13 @@ static void wxDo(Scheme_Object *proc, int argc, Scheme_Object **argv)
   /* wxDo might be called when MrEd is sleeping (i.e.,
      blocked on WNE in OS X). Since we're hijacking the
      thread, save an restore block information. */
-  block_descriptor = scheme_current_thread->block_descriptor;
-  scheme_current_thread->block_descriptor = 0;
+  block_descriptor = thread->block_descriptor;
+  thread->block_descriptor = 0;
 
   scheme_start_atomic();
 
-  save = scheme_current_thread->error_buf;
-  scheme_current_thread->error_buf = &newbuf;
+  save = thread->error_buf;
+  thread->error_buf = &newbuf;
 
   if (scheme_setjmp(newbuf)) {
     scheme_clear_escape();
@@ -3901,8 +3647,9 @@ static void wxDo(Scheme_Object *proc, int argc, Scheme_Object **argv)
     scheme_apply(proc, argc, argv);
   }
 
-  scheme_current_thread->error_buf = save;
-  scheme_current_thread->block_descriptor = block_descriptor;
+  thread = scheme_get_current_thread();
+  thread->error_buf = save;
+  thread->block_descriptor = block_descriptor;
 
   scheme_end_atomic_no_swap();
 }
@@ -3925,13 +3672,6 @@ void wxDrop_Runtime(char **argv, int argc)
 #if defined(wx_mac) || defined(wx_msw)
 void wxDrop_Quit()
 {
-#if WINDOW_STDIO
-  if (ioFrame) {
-    if (ioFrame->OnClose())
-      ioFrame->Show(FALSE);
-  }
-#endif
-
   wxDo(wxs_app_quit_proc, 0, NULL);
 }
 #endif
