@@ -1,6 +1,6 @@
 /*
   MzScheme
-  Copyright (c) 2004-2008 PLT Scheme Inc.
+  Copyright (c) 2004-2009 PLT Scheme Inc.
   Copyright (c) 1995-2001 Matthew Flatt
   All rights reserved.
 
@@ -166,6 +166,16 @@ typedef struct FSSpec mzFSSpec;
 
 #define MZ_EXTERN extern MZ_DLLSPEC
 
+#ifdef MZ_USE_PLACES
+# if _MSC_VER
+#  define THREAD_LOCAL __declspec(thread)
+# else
+#  define THREAD_LOCAL __thread
+# endif
+#else
+# define THREAD_LOCAL /* empty */
+#endif
+
 #if defined(MZ_USE_JIT_PPC) || defined(MZ_USE_JIT_I386) || defined(MZ_USE_JIT_X86_64)
 # define MZ_USE_JIT
 #endif
@@ -629,7 +639,7 @@ typedef struct Scheme_Offset_Cptr
 #define SCHEME_PRIM_IS_METHOD 4096
 #define SCHEME_PRIM_IS_CLOSURE 8192
 #define SCHEME_PRIM_IS_UNARY_INLINED 16384
-#define SCHEME_PRIM_IS_MIN_NARY_INLINED 32768
+#define SCHEME_PRIM_IS_NARY_INLINED 32768
 
 /* Values with SCHEME_PRIM_OPT_MASK, earlier implies later: */
 #define SCHEME_PRIM_OPT_FOLDING    3
@@ -780,7 +790,7 @@ typedef struct {
 
 typedef struct Scheme_Hash_Table
 {
-  Scheme_Inclhash_Object iso;
+  Scheme_Inclhash_Object iso; /* 0x1 flag => marshal as #t (hack for stxobj bytecode) */
   int size; /* power of 2 */
   int count;
   Scheme_Object **keys;
@@ -997,6 +1007,9 @@ typedef struct Scheme_Thread {
   char ran_some;
   char suspend_to_kill;
 
+  struct Scheme_Thread *return_marks_to;
+  Scheme_Object *returned_marks;
+
   struct Scheme_Overflow *overflow;
 
   struct Scheme_Comp_Env *current_local_env;
@@ -1011,6 +1024,7 @@ typedef struct Scheme_Thread {
   struct Scheme_Marshal_Tables *current_mt;
 
   Scheme_Object *constant_folding; /* compiler hack */
+  Scheme_Object *reading_delayed; /* reader hack */
 
   Scheme_Object *(*overflow_k)(void);
   Scheme_Object *overflow_reply;
@@ -1065,6 +1079,10 @@ typedef struct Scheme_Thread {
 
   /* save thread-specific GMP state: */
   long gmp_tls[6];
+  void *gmp_tls_data;
+
+  long accum_process_msec;
+  long current_start_process_msec;
 
   struct Scheme_Thread_Custodian_Hop *mr_hop;
   Scheme_Custodian_Reference *mref;
@@ -1078,7 +1096,7 @@ typedef struct Scheme_Thread {
   Scheme_Object *mbox_sema;
 
 #ifdef MZ_PRECISE_GC
-  struct gc_thread_info *gc_info; /* managed by the GC */
+  struct GC_Thread_Info *gc_info; /* managed by the GC */
 #endif
 } Scheme_Thread;
 
@@ -1090,19 +1108,37 @@ typedef struct Scheme_Thread {
 
 typedef void (*Scheme_Kill_Action_Func)(void *);
 
+#define ESCAPE_BLOCK(return_code) \
+    thread = scheme_get_current_thread(); \
+    savebuf = thread->error_buf; \
+    thread->error_buf = &newbuf; \
+    thread = NULL; \
+    if (scheme_setjmp(newbuf)) \
+    { \
+      thread = scheme_get_current_thread(); \
+      thread->error_buf = savebuf; \
+      scheme_clear_escape(); \
+      return return_code; \
+    }
+
 # define BEGIN_ESCAPEABLE(func, data) \
     { mz_jmp_buf * volatile savebuf, newbuf; \
+      Scheme_Thread *thread; \
+      thread = scheme_get_current_thread(); \
       scheme_push_kill_action((Scheme_Kill_Action_Func)func, (void *)data); \
-      savebuf = scheme_current_thread->error_buf; \
-      scheme_current_thread->error_buf = &newbuf; \
+      savebuf = thread->error_buf; \
+      thread->error_buf = &newbuf; \
+      thread = NULL; \
       if (scheme_setjmp(newbuf)) { \
         scheme_pop_kill_action(); \
         func(data); \
         scheme_longjmp(*savebuf, 1); \
       } else {
 # define END_ESCAPEABLE() \
+      thread = scheme_get_current_thread(); \
       scheme_pop_kill_action(); \
-      scheme_current_thread->error_buf = savebuf; } }
+      thread->error_buf = savebuf; \
+      thread = NULL; } }
 
 
 /*========================================================================*/
@@ -1166,6 +1202,7 @@ enum {
   MZCONFIG_ALLOW_SET_UNDEFINED,
   MZCONFIG_COMPILE_MODULE_CONSTS,
   MZCONFIG_USE_JIT,
+  MZCONFIG_DISALLOW_INLINE,
 
   MZCONFIG_CUSTODIAN,
   MZCONFIG_INSPECTOR,
@@ -1380,11 +1417,11 @@ typedef void (*Scheme_Invoke_Proc)(Scheme_Env *env, long phase_shift,
 
 #define SCHEME_ASSERT(expr,msg) ((expr) ? 1 : (scheme_signal_error(msg), 0))
 
+#ifndef MZ_USE_PLACES
 #define scheme_eval_wait_expr (scheme_current_thread->ku.eval.wait_expr)
 #define scheme_tail_rator (scheme_current_thread->ku.apply.tail_rator)
 #define scheme_tail_num_rands (scheme_current_thread->ku.apply.tail_num_rands)
 #define scheme_tail_rands (scheme_current_thread->ku.apply.tail_rands)
-#define scheme_overflow_k (scheme_current_thread->overflow_k)
 #define scheme_overflow_reply (scheme_current_thread->overflow_reply)
 
 #define scheme_error_buf *(scheme_current_thread->error_buf)
@@ -1392,6 +1429,7 @@ typedef void (*Scheme_Invoke_Proc)(Scheme_Env *env, long phase_shift,
 
 #define scheme_multiple_count (scheme_current_thread->ku.multiple.count)
 #define scheme_multiple_array (scheme_current_thread->ku.multiple.array)
+#endif
 
 #define scheme_setjmpup(b, base, s) scheme_setjmpup_relative(b, base, s, NULL)
 
@@ -1427,11 +1465,21 @@ typedef void (*Scheme_Invoke_Proc)(Scheme_Env *env, long phase_shift,
 #define _scheme_force_value(v) ((v == SCHEME_TAIL_CALL_WAITING) ? scheme_force_value(v) : v)
 
 #define scheme_tail_apply_buffer_wp(n, p) ((p)->tail_buffer)
-#define scheme_tail_apply_buffer(n) scheme_tail_apply_buffer_wp(n, scheme_current_thread)
+#define scheme_tail_apply_buffer(n) \
+{ \
+  Scheme_Thread *thread; \
+  thread = scheme_get_current_thread(); \
+  scheme_tail_apply_buffer_wp(n, thread);\
+}
 
 #define _scheme_tail_apply_no_copy_wp_tcw(f, n, args, p, tcw) (p->ku.apply.tail_rator = f, p->ku.apply.tail_rands = args, p->ku.apply.tail_num_rands = n, tcw)
 #define _scheme_tail_apply_no_copy_wp(f, n, args, p) _scheme_tail_apply_no_copy_wp_tcw(f, n, args, p, SCHEME_TAIL_CALL_WAITING)
-#define _scheme_tail_apply_no_copy(f, n, args) _scheme_tail_apply_no_copy_wp(f, n, args, scheme_current_thread)
+#define _scheme_tail_apply_no_copy(f, n, args) \
+{ \
+  Scheme_Thread *thread; \
+  thread = scheme_get_current_thread(); \
+  _scheme_tail_apply_no_copy_wp(f, n, args, thread) \
+}
 
 #define scheme_thread_block_w_thread(t,p) scheme_thread_block(t)
 
@@ -1648,9 +1696,15 @@ MZ_EXTERN void scheme_set_logging(int syslog_level, int stderr_level);
 
 MZ_EXTERN int scheme_get_allow_set_undefined();
 
+#ifndef MZ_USE_PLACES
+MZ_EXTERN THREAD_LOCAL Scheme_Thread *scheme_current_thread;
+MZ_EXTERN THREAD_LOCAL Scheme_Thread *scheme_first_thread;
+#endif
+MZ_EXTERN Scheme_Thread *scheme_get_current_thread();
+MZ_EXTERN long scheme_get_multiple_count();
+MZ_EXTERN Scheme_Object **scheme_get_multiple_array();
+MZ_EXTERN void scheme_set_current_thread_ran_some();
 
-MZ_EXTERN Scheme_Thread *scheme_current_thread;
-MZ_EXTERN Scheme_Thread *scheme_first_thread;
 
 /* Set these global hooks (optionally): */
 typedef void (*Scheme_Exit_Proc)(int v);
@@ -1689,6 +1743,7 @@ MZ_EXTERN void scheme_set_collects_path(Scheme_Object *p);
 MZ_EXTERN void scheme_set_original_dir(Scheme_Object *d);
 
 MZ_EXTERN void scheme_init_collection_paths(Scheme_Env *global_env, Scheme_Object *extra_dirs);
+MZ_EXTERN void scheme_init_collection_paths_post(Scheme_Env *global_env, Scheme_Object *extra_dirs, Scheme_Object *extra_post_dirs);
 
 /* Initialization */
 MZ_EXTERN Scheme_Env *scheme_basic_env(void);
@@ -1707,6 +1762,9 @@ MZ_EXTERN int scheme_get_external_event_fd(void);
 MZ_EXTERN void scheme_set_stack_base(void *base, int no_auto_statics);
 MZ_EXTERN void scheme_set_stack_bounds(void *base, void *deepest, int no_auto_statics);
 
+typedef void (*Scheme_Report_Out_Of_Memory_Proc)(void);
+MZ_EXTERN void scheme_set_report_out_of_memory(Scheme_Report_Out_Of_Memory_Proc p);
+
 /* Stack-preparation start-up: */
 typedef int (*Scheme_Nested_Main)(void *data);
 MZ_EXTERN int scheme_main_stack_setup(int no_auto_statics, Scheme_Nested_Main _main, void *data);
