@@ -1,6 +1,6 @@
 /*
   MzScheme
-  Copyright (c) 2004-2008 PLT Scheme Inc.
+  Copyright (c) 2004-2009 PLT Scheme Inc.
   Copyright (c) 1995-2001 Matthew Flatt
 
     This library is free software; you can redistribute it and/or
@@ -34,6 +34,21 @@
 
 int (*scheme_check_print_is_obj)(Scheme_Object *o);
 
+#define QUICK_ENCODE_BUFFER_SIZE 256
+static THREAD_LOCAL char *quick_buffer = NULL;
+static THREAD_LOCAL char *quick_encode_buffer = NULL;
+
+/* FIXME places possible race condition on growing printer size */
+static Scheme_Type_Printer *printers;
+static int printers_count;
+
+static Scheme_Hash_Table *cache_ht;
+
+/* read-only globals */
+static char compacts[_CPT_COUNT_];
+static Scheme_Hash_Table *global_constants_ht;
+static Scheme_Object *quote_link_symbol = NULL;
+
 /* Flag for debugging compiled code in printed form: */
 #define NO_COMPACT 0
 
@@ -97,26 +112,16 @@ static char *print_to_string(Scheme_Object *obj, long * volatile len, int write,
 static void custom_write_struct(Scheme_Object *s, Scheme_Hash_Table *ht, 
                                 Scheme_Marshal_Tables *mt,
 				PrintParams *pp, int notdisplay);
-static Scheme_Object *writable_struct_subs(Scheme_Object *s, PrintParams *pp);
+static Scheme_Object *writable_struct_subs(Scheme_Object *s, int for_write, PrintParams *pp);
 
-static Scheme_Object *quote_link_symbol = NULL;
-static char *quick_buffer = NULL;
-static char *quick_encode_buffer = NULL;
-
-static Scheme_Type_Printer *printers;
-static int printers_count;
-
-#define QUICK_ENCODE_BUFFER_SIZE 256
-
-static char compacts[_CPT_COUNT_];
-
-static Scheme_Hash_Table *global_constants_ht;
 
 #define print_compact(pp, v) print_this_string(pp, &compacts[v], 0, 1)
 
 #define PRINTABLE_STRUCT(obj, pp) (scheme_inspector_sees_part(obj, pp->inspector, -1))
 #define SCHEME_PREFABP(obj) (((Scheme_Structure *)(obj))->stype->prefab_key)
 
+#define SCHEME_HASHTPx(obj) ((SCHEME_HASHTP(obj) && !(MZ_OPT_HASH_KEY(&(((Scheme_Hash_Table *)obj)->iso)) & 0x1)))
+
 #define HAS_SUBSTRUCT(obj, qk) \
    (SCHEME_PAIRP(obj) \
     || SCHEME_MUTABLE_PAIRP(obj) \
@@ -126,23 +131,16 @@ static Scheme_Hash_Table *global_constants_ht;
 	   && SCHEME_STRUCTP(obj) \
 	   && PRINTABLE_STRUCT(obj, pp), 0)) \
     || (qk(SCHEME_STRUCTP(obj) && scheme_is_writable_struct(obj), 0)) \
-    || (qk(pp->print_hash_table, 1) && (SCHEME_HASHTP(obj) || SCHEME_HASHTRP(obj))))
+    || (qk(pp->print_hash_table, 1) && (SCHEME_HASHTPx(obj) || SCHEME_HASHTRP(obj))))
 #define ssQUICK(x, isbox) x
 #define ssQUICKp(x, isbox) (pp ? x : isbox)
 #define ssALL(x, isbox) 1
 #define ssALLp(x, isbox) isbox
 
-static Scheme_Hash_Table *cache_ht;
-
 void scheme_init_print(Scheme_Env *env)
 {
   int i;
 
-  REGISTER_SO(quick_buffer);
-  REGISTER_SO(quick_encode_buffer);
-  
-  quick_buffer = (char *)scheme_malloc_atomic(100);
-  quick_encode_buffer = (char *)scheme_malloc_atomic(QUICK_ENCODE_BUFFER_SIZE);
 
   REGISTER_SO(quote_link_symbol);
   
@@ -159,6 +157,15 @@ void scheme_init_print(Scheme_Env *env)
   REGISTER_SO(cache_ht);
 }
 
+void scheme_init_print_buffers_places() 
+{
+  REGISTER_SO(quick_buffer);
+  REGISTER_SO(quick_encode_buffer);
+  
+  quick_buffer = (char *)scheme_malloc_atomic(100);
+  quick_encode_buffer = (char *)scheme_malloc_atomic(QUICK_ENCODE_BUFFER_SIZE);
+}
+
 Scheme_Object *scheme_make_svector(mzshort c, mzshort *a)
 {
   Scheme_Object *o;
@@ -389,7 +396,7 @@ scheme_internal_print(Scheme_Object *obj, Scheme_Object *port)
 }
 
 #ifdef DO_STACK_CHECK
-static int check_cycles(Scheme_Object *, Scheme_Hash_Table *ht, PrintParams *);
+static int check_cycles(Scheme_Object *, int, Scheme_Hash_Table *ht, PrintParams *);
 
 static Scheme_Object *check_cycle_k(void)
 {
@@ -402,12 +409,12 @@ static Scheme_Object *check_cycle_k(void)
   p->ku.k.p2 = NULL;
   p->ku.k.p3 = NULL;
 
-  return check_cycles(o, ht, pp)
+  return check_cycles(o, p->ku.k.i1, ht, pp)
     ? scheme_true : scheme_false;
 }
 #endif
 
-static int check_cycles(Scheme_Object *obj, Scheme_Hash_Table *ht, PrintParams *pp)
+static int check_cycles(Scheme_Object *obj, int for_write, Scheme_Hash_Table *ht, PrintParams *pp)
 {
   Scheme_Type t;
 
@@ -419,6 +426,7 @@ static int check_cycles(Scheme_Object *obj, Scheme_Hash_Table *ht, PrintParams *
       scheme_current_thread->ku.k.p1 = (void *)obj;
       scheme_current_thread->ku.k.p2 = (void *)ht;
       scheme_current_thread->ku.k.p3 = (void *)pp;
+      scheme_current_thread->ku.k.i1 = for_write;
       return SCHEME_TRUEP(scheme_handle_stack_overflow(check_cycle_k));
     }
   }
@@ -446,27 +454,27 @@ static int check_cycles(Scheme_Object *obj, Scheme_Hash_Table *ht, PrintParams *
     return 0;
 
   if (SCHEME_PAIRP(obj) || SCHEME_MUTABLE_PAIRP(obj)) {
-    if (check_cycles(SCHEME_CAR(obj), ht, pp))
+    if (check_cycles(SCHEME_CAR(obj), for_write, ht, pp))
       return 1;
-    if (check_cycles(SCHEME_CDR(obj), ht, pp))
+    if (check_cycles(SCHEME_CDR(obj), for_write, ht, pp))
       return 1;
   } else if (SCHEME_BOXP(obj)) {
     /* got here => printable */
-    if (check_cycles(SCHEME_BOX_VAL(obj), ht, pp))
+    if (check_cycles(SCHEME_BOX_VAL(obj), for_write, ht, pp))
       return 1;
   } else if (SCHEME_VECTORP(obj)) {
     int i, len;
 
     len = SCHEME_VEC_SIZE(obj);
     for (i = 0; i < len; i++) {
-      if (check_cycles(SCHEME_VEC_ELS(obj)[i], ht, pp)) {
+      if (check_cycles(SCHEME_VEC_ELS(obj)[i], for_write, ht, pp)) {
 	return 1;
       }
     }
   } else if (SAME_TYPE(t, scheme_structure_type)
 	     || SAME_TYPE(t, scheme_proc_struct_type)) {
     if (scheme_is_writable_struct(obj)) {
-      if (check_cycles(writable_struct_subs(obj, pp), ht, pp))
+      if (check_cycles(writable_struct_subs(obj, for_write, pp), for_write, ht, pp))
 	return 1;
     } else {
       /* got here => printable */
@@ -474,13 +482,13 @@ static int check_cycles(Scheme_Object *obj, Scheme_Hash_Table *ht, PrintParams *
 
       while (i--) {
 	if (scheme_inspector_sees_part(obj, pp->inspector, i)) {
-	  if (check_cycles(((Scheme_Structure *)obj)->slots[i], ht, pp)) {
+	  if (check_cycles(((Scheme_Structure *)obj)->slots[i], for_write, ht, pp)) {
 	    return 1;
 	  }
 	}
       }
     }
-  } else if (SCHEME_HASHTP(obj)) {
+  } else if (SCHEME_HASHTPx(obj)) {
     /* got here => printable */
     Scheme_Hash_Table *t;
     Scheme_Object **keys, **vals, *val;
@@ -492,9 +500,9 @@ static int check_cycles(Scheme_Object *obj, Scheme_Hash_Table *ht, PrintParams *
     for (i = t->size; i--; ) {
       if (vals[i]) {
 	val = vals[i];
-	if (check_cycles(keys[i], ht, pp))
+	if (check_cycles(keys[i], for_write, ht, pp))
 	  return 1;
-	if (check_cycles(val, ht, pp))
+	if (check_cycles(val, for_write, ht, pp))
 	  return 1;
       }
     }
@@ -507,9 +515,9 @@ static int check_cycles(Scheme_Object *obj, Scheme_Hash_Table *ht, PrintParams *
     i = scheme_hash_tree_next(t, -1);
     while (i != -1) {
       scheme_hash_tree_index(t, i, &key, &val);
-      if (check_cycles(key, ht, pp))
+      if (check_cycles(key, for_write, ht, pp))
         return 1;
-      if (check_cycles(val, ht, pp))
+      if (check_cycles(val, for_write, ht, pp))
         return 1;
       i = scheme_hash_tree_next(t, i);
     }
@@ -529,9 +537,7 @@ START_XFORM_SKIP;
    version takes to long, we back out to the general case. (We don't
    even check for stack overflow, so keep the max limit low.) */
 
-static int fast_checker_counter;
-
-static int check_cycles_fast(Scheme_Object *obj, PrintParams *pp)
+static int check_cycles_fast(Scheme_Object *obj, PrintParams *pp, int *fast_checker_counter)
 {
   Scheme_Type t;
   int cycle = 0;
@@ -540,18 +546,18 @@ static int check_cycles_fast(Scheme_Object *obj, PrintParams *pp)
   if (t < 0)
     return 1;
 
-  if (fast_checker_counter-- < 0)
+  if ((*fast_checker_counter)-- < 0)
     return -1;
 
   if (SCHEME_PAIRP(obj) || SCHEME_MUTABLE_PAIRP(obj)) {
     obj->type = -t;
-    cycle = check_cycles_fast(SCHEME_CAR(obj), pp);
+    cycle = check_cycles_fast(SCHEME_CAR(obj), pp, fast_checker_counter);
     if (!cycle)
-      cycle = check_cycles_fast(SCHEME_CDR(obj), pp);
+      cycle = check_cycles_fast(SCHEME_CDR(obj), pp, fast_checker_counter);
     obj->type = t;
   } else if (pp->print_box && SCHEME_BOXP(obj)) {
     obj->type = -t;
-    cycle = check_cycles_fast(SCHEME_BOX_VAL(obj), pp);
+    cycle = check_cycles_fast(SCHEME_BOX_VAL(obj), pp, fast_checker_counter);
     obj->type = t;
   } else if (SCHEME_VECTORP(obj)) {
     int i, len;
@@ -559,7 +565,7 @@ static int check_cycles_fast(Scheme_Object *obj, PrintParams *pp)
     obj->type = -t;
     len = SCHEME_VEC_SIZE(obj);
     for (i = 0; i < len; i++) {
-      cycle = check_cycles_fast(SCHEME_VEC_ELS(obj)[i], pp);
+      cycle = check_cycles_fast(SCHEME_VEC_ELS(obj)[i], pp, fast_checker_counter);
       if (cycle)
 	break;
     }
@@ -578,7 +584,7 @@ static int check_cycles_fast(Scheme_Object *obj, PrintParams *pp)
       obj->type = -t;
       while (i--) {
 	if (scheme_inspector_sees_part(obj, pp->inspector, i)) {
-	  cycle = check_cycles_fast(((Scheme_Structure *)obj)->slots[i], pp);
+	  cycle = check_cycles_fast(((Scheme_Structure *)obj)->slots[i], pp, fast_checker_counter);
 	  if (cycle)
 	    break;
 	}
@@ -587,7 +593,7 @@ static int check_cycles_fast(Scheme_Object *obj, PrintParams *pp)
     } else
       cycle = 0;
   } else if (pp->print_hash_table
-	     && SCHEME_HASHTP(obj)) {
+	     && SCHEME_HASHTPx(obj)) {
     if (!((Scheme_Hash_Table *)obj)->count)
       cycle = 0;
     else
@@ -611,7 +617,7 @@ END_XFORM_SKIP;
 #endif
 
 #ifdef DO_STACK_CHECK
-static void setup_graph_table(Scheme_Object *obj, Scheme_Hash_Table *ht, int *counter, PrintParams *pp);
+static void setup_graph_table(Scheme_Object *obj, int for_write, Scheme_Hash_Table *ht, int *counter, PrintParams *pp);
 
 static Scheme_Object *setup_graph_k(void)
 {
@@ -620,19 +626,20 @@ static Scheme_Object *setup_graph_k(void)
   Scheme_Hash_Table *ht = (Scheme_Hash_Table *)p->ku.k.p2;
   int *counter = (int *)p->ku.k.p3;
   PrintParams *pp = (PrintParams *)p->ku.k.p4;
+  int for_write = p->ku.k.i1;
 
   p->ku.k.p1 = NULL;
   p->ku.k.p2 = NULL;
   p->ku.k.p3 = NULL;
   p->ku.k.p4 = NULL;
 
-  setup_graph_table(o, ht, counter, pp);
+  setup_graph_table(o, for_write, ht, counter, pp);
 
   return scheme_false;
 }
 #endif
 
-static void setup_graph_table(Scheme_Object *obj, Scheme_Hash_Table *ht,
+static void setup_graph_table(Scheme_Object *obj, int for_write, Scheme_Hash_Table *ht,
 			      int *counter, PrintParams *pp)
 {
   if (HAS_SUBSTRUCT(obj, ssQUICKp)) {
@@ -648,6 +655,7 @@ static void setup_graph_table(Scheme_Object *obj, Scheme_Hash_Table *ht,
 	scheme_current_thread->ku.k.p2 = (void *)ht;
 	scheme_current_thread->ku.k.p3 = (void *)counter;
 	scheme_current_thread->ku.k.p4 = (void *)pp;
+        scheme_current_thread->ku.k.i1 = for_write;
 	scheme_handle_stack_overflow(setup_graph_k);
 	return;
       }
@@ -671,32 +679,32 @@ static void setup_graph_table(Scheme_Object *obj, Scheme_Hash_Table *ht,
   SCHEME_USE_FUEL(1);
 
   if (SCHEME_PAIRP(obj) || SCHEME_MUTABLE_PAIRP(obj)) {
-    setup_graph_table(SCHEME_CAR(obj), ht, counter, pp);
-    setup_graph_table(SCHEME_CDR(obj), ht, counter, pp);
+    setup_graph_table(SCHEME_CAR(obj), for_write, ht, counter, pp);
+    setup_graph_table(SCHEME_CDR(obj), for_write, ht, counter, pp);
   } else if ((!pp || pp->print_box) && SCHEME_BOXP(obj)) {
-    setup_graph_table(SCHEME_BOX_VAL(obj), ht, counter, pp);
+    setup_graph_table(SCHEME_BOX_VAL(obj), for_write, ht, counter, pp);
   } else if (SCHEME_VECTORP(obj)) {
     int i, len;
 
     len = SCHEME_VEC_SIZE(obj);
     for (i = 0; i < len; i++) {
-      setup_graph_table(SCHEME_VEC_ELS(obj)[i], ht, counter, pp);
+      setup_graph_table(SCHEME_VEC_ELS(obj)[i], for_write, ht, counter, pp);
     }
   } else if (pp && SCHEME_STRUCTP(obj)) { /* got here => printable */
     if (scheme_is_writable_struct(obj)) {
       if (pp->print_unreadable) {
-	obj = writable_struct_subs(obj, pp);
-	setup_graph_table(obj, ht, counter, pp);
+	obj = writable_struct_subs(obj, for_write, pp);
+	setup_graph_table(obj, for_write, ht, counter, pp);
       }
     } else {
       int i = SCHEME_STRUCT_NUM_SLOTS(obj);
 
       while (i--) {
 	if (scheme_inspector_sees_part(obj, pp->inspector, i))
-	  setup_graph_table(((Scheme_Structure *)obj)->slots[i], ht, counter, pp);
+	  setup_graph_table(((Scheme_Structure *)obj)->slots[i], for_write, ht, counter, pp);
       }
     }
-  } else if (pp && SCHEME_HASHTP(obj)) { /* got here => printable */
+  } else if (pp && SCHEME_HASHTPx(obj)) { /* got here => printable */
     Scheme_Hash_Table *t;
     Scheme_Object **keys, **vals, *val;
     int i;
@@ -707,8 +715,8 @@ static void setup_graph_table(Scheme_Object *obj, Scheme_Hash_Table *ht,
     for (i = t->size; i--; ) {
       if (vals[i]) {
 	val = vals[i];
-	setup_graph_table(keys[i], ht, counter, pp);
-	setup_graph_table(val, ht, counter, pp);
+	setup_graph_table(keys[i], for_write, ht, counter, pp);
+	setup_graph_table(val, for_write, ht, counter, pp);
       }
     }
   } else if (SCHEME_HASHTRP(obj)) {
@@ -720,8 +728,8 @@ static void setup_graph_table(Scheme_Object *obj, Scheme_Hash_Table *ht,
     i = scheme_hash_tree_next(t, -1);
     while (i != -1) {
       scheme_hash_tree_index(t, i, &key, &val);
-      setup_graph_table(key, ht, counter, pp);
-      setup_graph_table(val, ht, counter, pp);
+      setup_graph_table(key, for_write, ht, counter, pp);
+      setup_graph_table(val, for_write, ht, counter, pp);
       i = scheme_hash_tree_next(t, i);
     }
   }
@@ -729,7 +737,7 @@ static void setup_graph_table(Scheme_Object *obj, Scheme_Hash_Table *ht,
 
 #define CACHE_HT_SIZE_LIMIT 32
 
-Scheme_Hash_Table *scheme_setup_datum_graph(Scheme_Object *o, void *for_print)
+static Scheme_Hash_Table *setup_datum_graph(Scheme_Object *o, int for_write, void *for_print)
 {
   Scheme_Hash_Table *ht;
   int counter = 1;
@@ -740,7 +748,7 @@ Scheme_Hash_Table *scheme_setup_datum_graph(Scheme_Object *o, void *for_print)
   } else
     ht = scheme_make_hash_table(SCHEME_hash_ptr);
 
-  setup_graph_table(o, ht, &counter, (PrintParams *)for_print);
+  setup_graph_table(o, for_write, ht, &counter, (PrintParams *)for_print);
 
   if (counter > 1)
     return ht;
@@ -840,16 +848,16 @@ print_to_string(Scheme_Object *obj,
   if (params.print_graph)
     cycles = 1;
   else {
-    fast_checker_counter = 50;
-    cycles = check_cycles_fast(obj, (PrintParams *)&params);
+    int fast_checker_counter = 50;
+    cycles = check_cycles_fast(obj, (PrintParams *)&params, &fast_checker_counter);
     if (cycles == -1) {
       ht = scheme_make_hash_table(SCHEME_hash_ptr);
-      cycles = check_cycles(obj, ht, (PrintParams *)&params);
+      cycles = check_cycles(obj, write, ht, (PrintParams *)&params);
     }
   }
 
   if (cycles)
-    ht = scheme_setup_datum_graph(obj, (PrintParams *)&params);
+    ht = setup_datum_graph(obj, write, (PrintParams *)&params);
   else
     ht = NULL;
 
@@ -1053,8 +1061,9 @@ static void do_print_string(int compact, int notdisplay,
     buf = (char *)scheme_malloc_atomic(el);
     reset = 0;
   }
-  el = scheme_utf8_encode(s, offset, offset + l, 
-			  (unsigned char *)buf, 0, 0);
+
+  el = scheme_utf8_encode(s, offset, offset + l, (unsigned char *)buf, 0, 0);
+
   if (compact) {
     print_compact(pp, CPT_CHAR_STRING);
     print_compact_number(pp, el);
@@ -1063,6 +1072,7 @@ static void do_print_string(int compact, int notdisplay,
   } else {
     print_char_string(buf, el, s, offset, l, notdisplay, 0, pp);
   }
+
   if (reset)
     quick_encode_buffer = buf;
 }
@@ -1823,7 +1833,7 @@ print(Scheme_Object *obj, int notdisplay, int compact, Scheme_Hash_Table *ht,
       }
     }
   else if ((compact || pp->print_hash_table) 
-           && (SCHEME_HASHTP(obj) || SCHEME_HASHTRP(obj)))
+           && (SCHEME_HASHTPx(obj) || SCHEME_HASHTRP(obj)))
     {
       Scheme_Hash_Table *t;
       Scheme_Hash_Tree *tr;
@@ -1834,18 +1844,29 @@ print(Scheme_Object *obj, int notdisplay, int compact, Scheme_Hash_Table *ht,
 	print_compact(pp, CPT_HASH_TABLE);
 	if ((SCHEME_HASHTP(obj) && scheme_is_hash_table_equal(obj))
             || (SCHEME_HASHTRP(obj) && scheme_is_hash_tree_equal(obj)))
-	  print_compact_number(pp, 0);
-	else
 	  print_compact_number(pp, 1);
+	else if ((SCHEME_HASHTP(obj) && scheme_is_hash_table_eqv(obj))
+                 || (SCHEME_HASHTRP(obj) && scheme_is_hash_tree_eqv(obj)))
+	  print_compact_number(pp, 2);
+	else
+	  print_compact_number(pp, 0);
       } else {
 	always_scheme(pp, 1);
 	print_utf8_string(pp, "#hash", 0, 5);
         if (SCHEME_HASHTP(obj)) {
-          if (!scheme_is_hash_table_equal(obj))
-            print_utf8_string(pp, "eq", 0, 2);
+          if (!scheme_is_hash_table_equal(obj)) {
+            if (scheme_is_hash_table_eqv(obj))
+              print_utf8_string(pp, "eqv", 0, 3);
+            else
+              print_utf8_string(pp, "eq", 0, 2);
+          }
         } else {
-          if (!scheme_is_hash_tree_equal(obj))
-            print_utf8_string(pp, "eq", 0, 2);
+          if (!scheme_is_hash_tree_equal(obj)) {
+            if (scheme_is_hash_tree_eqv(obj))
+              print_utf8_string(pp, "eqv", 0, 3);
+            else
+              print_utf8_string(pp, "eq", 0, 2);
+          }
         }
 	print_utf8_string(pp, "(", 0, 1);
       }
@@ -1899,6 +1920,12 @@ print(Scheme_Object *obj, int notdisplay, int compact, Scheme_Hash_Table *ht,
 
       closed = 1;
     }
+  else if (compact && SCHEME_HASHTP(obj))
+    {
+      /* since previous case didn't catch this table, it has a 0x1 flag
+         and should be marshalled as #t */
+      print_compact(pp, CPT_TRUE);
+    }
   else if (SAME_OBJ(obj, scheme_true))
     {
       if (compact)
@@ -1958,20 +1985,28 @@ print(Scheme_Object *obj, int notdisplay, int compact, Scheme_Hash_Table *ht,
 	    src = obj;
 
 	  if (SAME_OBJ(src, obj)) {
+            int l;
+            const char *s;
+            Scheme_Object *name;
+
 	    print_utf8_string(pp, "#<", 0, 2); /* used to have "struct:" prefix */
-	    {
-	      int l;
-	      const char *s;
-	      Scheme_Object *name = SCHEME_STRUCT_NAME_SYM(obj);
-	      
-	      s = scheme_symbol_name_and_size(name, (unsigned int *)&l, 
-					      (pp->print_struct
-					       ? SCHEME_SNF_FOR_TS
-					       : (pp->can_read_pipe_quote 
-						  ? SCHEME_SNF_PIPE_QUOTE
-						  : SCHEME_SNF_NO_PIPE_QUOTE)));
-	      print_utf8_string(pp, s, 0, l);
-	    }
+            if (scheme_reduced_procedure_struct
+                && scheme_is_struct_instance(scheme_reduced_procedure_struct, obj)) {
+              /* Since scheme_proc_struct_name_source() didn't redirect, this one
+                 must have a name. */
+              print_utf8_string(pp, "procedure:", 0, 10);
+              name = ((Scheme_Structure *)obj)->slots[2];
+            } else {
+	      name = SCHEME_STRUCT_NAME_SYM(obj);
+            }
+
+            s = scheme_symbol_name_and_size(name, (unsigned int *)&l, 
+                                            (pp->print_struct
+                                             ? SCHEME_SNF_FOR_TS
+                                             : (pp->can_read_pipe_quote 
+                                                ? SCHEME_SNF_PIPE_QUOTE
+                                                : SCHEME_SNF_NO_PIPE_QUOTE)));
+            print_utf8_string(pp, s, 0, l);
 	    PRINTADDRESS(pp, obj);
 	    print_utf8_string(pp, ">", 0, 1);
 	  } else {
@@ -2339,8 +2374,7 @@ print(Scheme_Object *obj, int notdisplay, int compact, Scheme_Hash_Table *ht,
         symtab_set(pp, mt, obj);
       }
     }
-  else if (compact && SAME_TYPE(SCHEME_TYPE(obj), scheme_module_variable_type)
-	   && !((Module_Variable *)obj)->mod_phase)
+  else if (compact && SAME_TYPE(SCHEME_TYPE(obj), scheme_module_variable_type))
     {
       Scheme_Object *idx;
 
@@ -2359,7 +2393,11 @@ print(Scheme_Object *obj, int notdisplay, int compact, Scheme_Hash_Table *ht,
           print(mv->modidx, notdisplay, 1, ht, mt, pp);
         }
 	print(mv->sym, notdisplay, 1, ht, mt, pp);
-	print_compact_number(pp, mv->pos);
+        if (((Module_Variable *)obj)->mod_phase) {
+          /* mod_phase must be 1 */
+          print_compact_number(pp, -2);
+        }
+        print_compact_number(pp, mv->pos);
 
         symtab_set(pp, mt, obj);
       }
@@ -2483,7 +2521,7 @@ print(Scheme_Object *obj, int notdisplay, int compact, Scheme_Hash_Table *ht,
       pp->print_box = 1;
       
       q_ht = scheme_make_hash_table(SCHEME_hash_ptr);
-      setup_graph_table(v, q_ht, &counter, pp);
+      setup_graph_table(v, notdisplay, q_ht, &counter, pp);
 
       if (compact)
 	print_compact(pp, CPT_QUOTE);
@@ -3246,7 +3284,7 @@ static Scheme_Object *accum_write(void *_b, int argc, Scheme_Object **argv)
   return scheme_void;
 }
 
-static Scheme_Object *writable_struct_subs(Scheme_Object *s, PrintParams *pp)
+static Scheme_Object *writable_struct_subs(Scheme_Object *s, int for_write, PrintParams *pp)
 {
   Scheme_Object *v, *o, *a[3], *b, *accum_proc;
   Scheme_Output_Port *op;
@@ -3270,7 +3308,7 @@ static Scheme_Object *writable_struct_subs(Scheme_Object *s, PrintParams *pp)
 
   a[0] = s;
   a[1] = o;
-  a[2] = scheme_false;
+  a[2] = (for_write ? scheme_true : scheme_false);
 
   scheme_apply_multi(v, 3, a);
 
