@@ -1,6 +1,6 @@
 /*
   MzScheme
-  Copyright (c) 2004-2008 PLT Scheme Inc.
+  Copyright (c) 2004-2009 PLT Scheme Inc.
   Copyright (c) 1995-2001 Matthew Flatt
   All rights reserved.
 
@@ -54,7 +54,9 @@ Scheme_Object *(*scheme_current_break_cell)();
 /*                                threads                                 */
 /*========================================================================*/
 #ifndef LINK_EXTENSIONS_BY_TABLE
+# ifndef MZ_USE_PLACES
 Scheme_Thread *scheme_current_thread;
+# endif
 volatile int scheme_fuel_counter;
 #else
 Scheme_Thread **scheme_current_thread_ptr;
@@ -64,6 +66,8 @@ Scheme_Thread *(*scheme_get_current_thread)();
 void (*scheme_start_atomic)(void);
 void (*scheme_end_atomic)(void);
 void (*scheme_end_atomic_no_swap)(void);
+void (*scheme_start_in_scheduler)(void);
+void (*scheme_end_in_scheduler)(void);
 void (*scheme_out_of_fuel)(void);
 Scheme_Object *(*scheme_thread)(Scheme_Object *thunk);
 Scheme_Object *(*scheme_thread_w_details)(Scheme_Object *thunk, 
@@ -93,6 +97,9 @@ void (*scheme_wait_input_allowed)(Scheme_Input_Port *port, int nonblock);
 int (*scheme_unless_ready)(Scheme_Object *unless);
 int (*scheme_in_main_thread)(void);
 void (*scheme_cancel_sleep)(void);
+void (*scheme_start_sleeper_thread)(void (*mzsleep)(float seconds, void *fds), float secs, void *fds, int hit_fd);
+void (*scheme_end_sleeper_thread)();
+void (*scheme_notify_sleep_progress)();
 Scheme_Object *(*scheme_make_thread_cell)(Scheme_Object *def_val, int inherited);
 Scheme_Object *(*scheme_thread_cell_get)(Scheme_Object *cell, Scheme_Thread_Cell_Table *cells);
 void (*scheme_thread_cell_set)(Scheme_Object *cell, Scheme_Thread_Cell_Table *cells, Scheme_Object *v);
@@ -142,6 +149,8 @@ int (*scheme_log_level_p)(Scheme_Logger *logger, int level);
 void (*scheme_log)(Scheme_Logger *logger, int level, int flags,
                           char *msg, ...);
 void (*scheme_log_message)(Scheme_Logger *logger, int level, char *buffer, long len, Scheme_Object *data);
+void (*scheme_log_abort)(char *buffer);
+void (*scheme_out_of_memory_abort)();
 void (*scheme_wrong_count)(const char *name, int minc, int maxc,
 				  int argc, Scheme_Object **argv);
 void (*scheme_wrong_count_m)(const char *name, int minc, int maxc,
@@ -289,6 +298,11 @@ void *(*GC_malloc_uncollectable)(size_t size_in_bytes);
 #  endif
 # endif
 #endif
+void *(*scheme_malloc_code)(long size);
+void (*scheme_free_code)(void *p);
+#ifndef MZ_PRECISE_GC
+void *(*scheme_malloc_gcable_code)(long size);
+#endif
 void *(*scheme_malloc_eternal)(size_t n);
 void (*scheme_end_stubborn_change)(void *p);
 void *(*scheme_calloc)(size_t num, size_t size);
@@ -320,6 +334,7 @@ void (*GC_register_traversers)(short tag, Size_Proc size, Mark_Proc mark, Fixup_
 void *(*GC_resolve)(void *p);
 void (*GC_mark)(const void *p);
 void (*GC_fixup)(void *p);
+void *(*GC_fixup_self)(void *p);
 #endif
 void **(*scheme_malloc_immobile_box)(void *p);
 void (*scheme_free_immobile_box)(void **b);
@@ -335,6 +350,7 @@ int (*scheme_bucket_table_equal)(Scheme_Bucket_Table *t1, Scheme_Bucket_Table *t
 Scheme_Bucket_Table *(*scheme_clone_bucket_table)(Scheme_Bucket_Table *bt);
 Scheme_Hash_Table *(*scheme_make_hash_table)(int type);
 Scheme_Hash_Table *(*scheme_make_hash_table_equal)();
+Scheme_Hash_Table *(*scheme_make_hash_table_eqv)();
 void (*scheme_hash_set)(Scheme_Hash_Table *table, Scheme_Object *key, Scheme_Object *val);
 Scheme_Object *(*scheme_hash_get)(Scheme_Hash_Table *table, Scheme_Object *key);
 Scheme_Object *(*scheme_eq_hash_get)(Scheme_Hash_Table *table, Scheme_Object *key);
@@ -342,14 +358,16 @@ void (*scheme_hash_set_atomic)(Scheme_Hash_Table *table, Scheme_Object *key, Sch
 Scheme_Object *(*scheme_hash_get_atomic)(Scheme_Hash_Table *table, Scheme_Object *key);
 int (*scheme_hash_table_equal)(Scheme_Hash_Table *t1, Scheme_Hash_Table *t2);
 int (*scheme_is_hash_table_equal)(Scheme_Object *o);
+int (*scheme_is_hash_table_eqv)(Scheme_Object *o);
 Scheme_Hash_Table *(*scheme_clone_hash_table)(Scheme_Hash_Table *bt);
-Scheme_Hash_Tree *(*scheme_make_hash_tree)(int eql);
+Scheme_Hash_Tree *(*scheme_make_hash_tree)(int kind);
 Scheme_Hash_Tree *(*scheme_hash_tree_set)(Scheme_Hash_Tree *tree, Scheme_Object *key, Scheme_Object *val);
 Scheme_Object *(*scheme_hash_tree_get)(Scheme_Hash_Tree *tree, Scheme_Object *key);
 long (*scheme_hash_tree_next)(Scheme_Hash_Tree *tree, long pos);
 int (*scheme_hash_tree_index)(Scheme_Hash_Tree *tree, long pos, Scheme_Object **_key, Scheme_Object **_val);
 int (*scheme_hash_tree_equal)(Scheme_Hash_Tree *t1, Scheme_Hash_Tree *t2);
 int (*scheme_is_hash_tree_equal)(Scheme_Object *o);
+int (*scheme_is_hash_tree_eqv)(Scheme_Object *o);
 /*========================================================================*/
 /*                   basic Scheme value constructors                      */
 /*========================================================================*/
@@ -407,7 +425,7 @@ Scheme_Object *(*scheme_make_sized_byte_string)(char *chars, long len, int copy)
 Scheme_Object *(*scheme_make_sized_offset_byte_string)(char *chars, long d, long len, int copy);
 Scheme_Object *(*scheme_make_immutable_sized_byte_string)(char *chars, long len, int copy);
 Scheme_Object *(*scheme_make_byte_string_without_copying)(char *chars);
-Scheme_Object *(*scheme_alloc_byte_string)(int size, char fill);
+Scheme_Object *(*scheme_alloc_byte_string)(long size, char fill);
 Scheme_Object *(*scheme_append_byte_string)(Scheme_Object *, Scheme_Object *);
 Scheme_Object *(*scheme_make_utf8_string)(const char *chars);
 Scheme_Object *(*scheme_make_sized_utf8_string)(char *chars, long len);
@@ -425,10 +443,10 @@ Scheme_Object *(*scheme_make_sized_char_string)(mzchar *chars, long len, int cop
 Scheme_Object *(*scheme_make_sized_offset_char_string)(mzchar *chars, long d, long len, int copy);
 Scheme_Object *(*scheme_make_immutable_sized_char_string)(mzchar *chars, long len, int copy);
 Scheme_Object *(*scheme_make_char_string_without_copying)(mzchar *chars);
-Scheme_Object *(*scheme_alloc_char_string)(int size, mzchar fill);
+Scheme_Object *(*scheme_alloc_char_string)(long size, mzchar fill);
 Scheme_Object *(*scheme_append_char_string)(Scheme_Object *, Scheme_Object *);
 mzchar *(*scheme_string_recase)(mzchar *s, int d, int len, int mode, int inplace, int *_len);
-Scheme_Object *(*scheme_make_vector)(int size, Scheme_Object *fill);
+Scheme_Object *(*scheme_make_vector)(long size, Scheme_Object *fill);
 Scheme_Object *(*scheme_make_integer_value)(long i);
 Scheme_Object *(*scheme_make_integer_value_from_unsigned)(unsigned long i);
 Scheme_Object *(*scheme_make_integer_value_from_long_long)(mzlonglong i);
@@ -718,6 +736,7 @@ void (*scheme_fdclr)(void *fd, int pos);
 int (*scheme_fdisset)(void *fd, int pos);
 void (*scheme_add_fd_handle)(void *h, void *fds, int repost);
 void (*scheme_add_fd_eventmask)(void *fds, int mask);
+void (*scheme_collapse_win_fd)(void *fds);
 void (*scheme_security_check_file)(const char *who, const char *filename, int guards);
 void (*scheme_security_check_file_link)(const char *who, const char *filename, const char *content);
 void (*scheme_security_check_network)(const char *who, const char *host, int port, int client);
@@ -743,7 +762,6 @@ Scheme_Object *(*scheme_read_byte_string)(Scheme_Object *port);
 /*========================================================================*/
 Scheme_Object *(*scheme_make_namespace)(int argc, Scheme_Object *argv[]);
 void (*scheme_add_namespace_option)(Scheme_Object *key, void (*f)(Scheme_Env *));
-void (*scheme_require_from_original_env)(Scheme_Env *env, int syntax_only);
 void (*scheme_add_global)(const char *name, Scheme_Object *val, Scheme_Env *env);
 void (*scheme_add_global_symbol)(Scheme_Object *name, Scheme_Object *val,
 			      Scheme_Env *env);
@@ -832,6 +850,8 @@ long (*scheme_equal_hash_key)(Scheme_Object *o);
 long (*scheme_equal_hash_key2)(Scheme_Object *o);
 long (*scheme_recur_equal_hash_key)(Scheme_Object *o, void *cycle_data);
 long (*scheme_recur_equal_hash_key2)(Scheme_Object *o, void *cycle_data);
+long (*scheme_eqv_hash_key)(Scheme_Object *o);
+long (*scheme_eqv_hash_key2)(Scheme_Object *o);
 void (*scheme_set_type_equality)(Scheme_Type type, 
                                         Scheme_Equal_Proc f,
                                         Scheme_Primary_Hash_Proc hash1,
@@ -867,6 +887,7 @@ long (*scheme_get_seconds)(void);
 long (*scheme_get_milliseconds)(void);
 double (*scheme_get_inexact_milliseconds)(void);
 long (*scheme_get_process_milliseconds)(void);
+long (*scheme_get_thread_milliseconds)(Scheme_Object *thrd);
 char *(*scheme_banner)(void);
 char *(*scheme_version)(void);
 int (*scheme_check_proc_arity)(const char *where, int a,
@@ -876,7 +897,6 @@ int (*scheme_check_proc_arity2)(const char *where, int a,
 				       int false_ok);
 char *(*scheme_make_provided_string)(Scheme_Object *o, int count, int *len);
 char *(*scheme_make_args_string)(char *s, int which, int argc, Scheme_Object **argv, long *len);
-void (*scheme_no_dumps)(char *why);
 const char *(*scheme_system_library_subpath)();
 void (*scheme_signal_received)(void);
 int (*scheme_char_strlen)(const mzchar *s);
