@@ -19,7 +19,9 @@
 #define scheme_inherit_cells (scheme_extension_table->scheme_inherit_cells)
 #define scheme_current_break_cell (scheme_extension_table->scheme_current_break_cell)
 #ifndef LINK_EXTENSIONS_BY_TABLE
+# ifndef MZ_USE_PLACES
 #define scheme_current_thread (scheme_extension_table->scheme_current_thread)
+# endif
 #define scheme_fuel_counter (scheme_extension_table->scheme_fuel_counter)
 #else
 #define scheme_current_thread_ptr (scheme_extension_table->scheme_current_thread_ptr)
@@ -29,6 +31,8 @@
 #define scheme_start_atomic (scheme_extension_table->scheme_start_atomic)
 #define scheme_end_atomic (scheme_extension_table->scheme_end_atomic)
 #define scheme_end_atomic_no_swap (scheme_extension_table->scheme_end_atomic_no_swap)
+#define scheme_start_in_scheduler (scheme_extension_table->scheme_start_in_scheduler)
+#define scheme_end_in_scheduler (scheme_extension_table->scheme_end_in_scheduler)
 #define scheme_out_of_fuel (scheme_extension_table->scheme_out_of_fuel)
 #define scheme_thread (scheme_extension_table->scheme_thread)
 #define scheme_thread_w_details (scheme_extension_table->scheme_thread_w_details)
@@ -49,6 +53,9 @@
 #define scheme_unless_ready (scheme_extension_table->scheme_unless_ready)
 #define scheme_in_main_thread (scheme_extension_table->scheme_in_main_thread)
 #define scheme_cancel_sleep (scheme_extension_table->scheme_cancel_sleep)
+#define scheme_start_sleeper_thread (scheme_extension_table->scheme_start_sleeper_thread)
+#define scheme_end_sleeper_thread (scheme_extension_table->scheme_end_sleeper_thread)
+#define scheme_notify_sleep_progress (scheme_extension_table->scheme_notify_sleep_progress)
 #define scheme_make_thread_cell (scheme_extension_table->scheme_make_thread_cell)
 #define scheme_thread_cell_get (scheme_extension_table->scheme_thread_cell_get)
 #define scheme_thread_cell_set (scheme_extension_table->scheme_thread_cell_set)
@@ -86,6 +93,8 @@
 #define scheme_log_level_p (scheme_extension_table->scheme_log_level_p)
 #define scheme_log (scheme_extension_table->scheme_log)
 #define scheme_log_message (scheme_extension_table->scheme_log_message)
+#define scheme_log_abort (scheme_extension_table->scheme_log_abort)
+#define scheme_out_of_memory_abort (scheme_extension_table->scheme_out_of_memory_abort)
 #define scheme_wrong_count (scheme_extension_table->scheme_wrong_count)
 #define scheme_wrong_count_m (scheme_extension_table->scheme_wrong_count_m)
 #define scheme_case_lambda_wrong_count (scheme_extension_table->scheme_case_lambda_wrong_count)
@@ -192,6 +201,11 @@
 #  endif
 # endif
 #endif
+#define scheme_malloc_code (scheme_extension_table->scheme_malloc_code)
+#define scheme_free_code (scheme_extension_table->scheme_free_code)
+#ifndef MZ_PRECISE_GC
+#define scheme_malloc_gcable_code (scheme_extension_table->scheme_malloc_gcable_code)
+#endif
 #define scheme_malloc_eternal (scheme_extension_table->scheme_malloc_eternal)
 #define scheme_end_stubborn_change (scheme_extension_table->scheme_end_stubborn_change)
 #define scheme_calloc (scheme_extension_table->scheme_calloc)
@@ -219,6 +233,7 @@
 #define GC_resolve (scheme_extension_table->GC_resolve)
 #define GC_mark (scheme_extension_table->GC_mark)
 #define GC_fixup (scheme_extension_table->GC_fixup)
+#define GC_fixup_self (scheme_extension_table->GC_fixup_self)
 #endif
 #define scheme_malloc_immobile_box (scheme_extension_table->scheme_malloc_immobile_box)
 #define scheme_free_immobile_box (scheme_extension_table->scheme_free_immobile_box)
@@ -231,6 +246,7 @@
 #define scheme_clone_bucket_table (scheme_extension_table->scheme_clone_bucket_table)
 #define scheme_make_hash_table (scheme_extension_table->scheme_make_hash_table)
 #define scheme_make_hash_table_equal (scheme_extension_table->scheme_make_hash_table_equal)
+#define scheme_make_hash_table_eqv (scheme_extension_table->scheme_make_hash_table_eqv)
 #define scheme_hash_set (scheme_extension_table->scheme_hash_set)
 #define scheme_hash_get (scheme_extension_table->scheme_hash_get)
 #define scheme_eq_hash_get (scheme_extension_table->scheme_eq_hash_get)
@@ -238,6 +254,7 @@
 #define scheme_hash_get_atomic (scheme_extension_table->scheme_hash_get_atomic)
 #define scheme_hash_table_equal (scheme_extension_table->scheme_hash_table_equal)
 #define scheme_is_hash_table_equal (scheme_extension_table->scheme_is_hash_table_equal)
+#define scheme_is_hash_table_eqv (scheme_extension_table->scheme_is_hash_table_eqv)
 #define scheme_clone_hash_table (scheme_extension_table->scheme_clone_hash_table)
 #define scheme_make_hash_tree (scheme_extension_table->scheme_make_hash_tree)
 #define scheme_hash_tree_set (scheme_extension_table->scheme_hash_tree_set)
@@ -246,6 +263,7 @@
 #define scheme_hash_tree_index (scheme_extension_table->scheme_hash_tree_index)
 #define scheme_hash_tree_equal (scheme_extension_table->scheme_hash_tree_equal)
 #define scheme_is_hash_tree_equal (scheme_extension_table->scheme_is_hash_tree_equal)
+#define scheme_is_hash_tree_eqv (scheme_extension_table->scheme_is_hash_tree_eqv)
 #define scheme_make_prim (scheme_extension_table->scheme_make_prim)
 #define scheme_make_noneternal_prim (scheme_extension_table->scheme_make_noneternal_prim)
 #define scheme_make_prim_w_arity (scheme_extension_table->scheme_make_prim_w_arity)
@@ -495,6 +513,7 @@
 #define scheme_fdisset (scheme_extension_table->scheme_fdisset)
 #define scheme_add_fd_handle (scheme_extension_table->scheme_add_fd_handle)
 #define scheme_add_fd_eventmask (scheme_extension_table->scheme_add_fd_eventmask)
+#define scheme_collapse_win_fd (scheme_extension_table->scheme_collapse_win_fd)
 #define scheme_security_check_file (scheme_extension_table->scheme_security_check_file)
 #define scheme_security_check_file_link (scheme_extension_table->scheme_security_check_file_link)
 #define scheme_security_check_network (scheme_extension_table->scheme_security_check_network)
@@ -513,7 +532,6 @@
 #define scheme_read_byte_string (scheme_extension_table->scheme_read_byte_string)
 #define scheme_make_namespace (scheme_extension_table->scheme_make_namespace)
 #define scheme_add_namespace_option (scheme_extension_table->scheme_add_namespace_option)
-#define scheme_require_from_original_env (scheme_extension_table->scheme_require_from_original_env)
 #define scheme_add_global (scheme_extension_table->scheme_add_global)
 #define scheme_add_global_symbol (scheme_extension_table->scheme_add_global_symbol)
 #define scheme_make_envunbox (scheme_extension_table->scheme_make_envunbox)
@@ -570,6 +588,8 @@
 #define scheme_equal_hash_key2 (scheme_extension_table->scheme_equal_hash_key2)
 #define scheme_recur_equal_hash_key (scheme_extension_table->scheme_recur_equal_hash_key)
 #define scheme_recur_equal_hash_key2 (scheme_extension_table->scheme_recur_equal_hash_key2)
+#define scheme_eqv_hash_key (scheme_extension_table->scheme_eqv_hash_key)
+#define scheme_eqv_hash_key2 (scheme_extension_table->scheme_eqv_hash_key2)
 #define scheme_set_type_equality (scheme_extension_table->scheme_set_type_equality)
 #define scheme_recur_equal (scheme_extension_table->scheme_recur_equal)
 #define scheme_build_list (scheme_extension_table->scheme_build_list)
@@ -601,13 +621,13 @@
 #define scheme_get_milliseconds (scheme_extension_table->scheme_get_milliseconds)
 #define scheme_get_inexact_milliseconds (scheme_extension_table->scheme_get_inexact_milliseconds)
 #define scheme_get_process_milliseconds (scheme_extension_table->scheme_get_process_milliseconds)
+#define scheme_get_thread_milliseconds (scheme_extension_table->scheme_get_thread_milliseconds)
 #define scheme_banner (scheme_extension_table->scheme_banner)
 #define scheme_version (scheme_extension_table->scheme_version)
 #define scheme_check_proc_arity (scheme_extension_table->scheme_check_proc_arity)
 #define scheme_check_proc_arity2 (scheme_extension_table->scheme_check_proc_arity2)
 #define scheme_make_provided_string (scheme_extension_table->scheme_make_provided_string)
 #define scheme_make_args_string (scheme_extension_table->scheme_make_args_string)
-#define scheme_no_dumps (scheme_extension_table->scheme_no_dumps)
 #define scheme_system_library_subpath (scheme_extension_table->scheme_system_library_subpath)
 #define scheme_signal_received (scheme_extension_table->scheme_signal_received)
 #define scheme_char_strlen (scheme_extension_table->scheme_char_strlen)
