@@ -1,6 +1,6 @@
 /*
   MzScheme
-  Copyright (c) 2004-2008 PLT Scheme Inc.
+  Copyright (c) 2004-2009 PLT Scheme Inc.
   Copyright (c) 1995-2001 Matthew Flatt
 
     This library is free software; you can redistribute it and/or
@@ -51,7 +51,7 @@ static void init_type_arrays()
   REGISTER_SO(scheme_type_hash2s);
   
   maxtype = _scheme_last_type_;
-  allocmax = maxtype + 10;
+  allocmax = maxtype + 100;
 
   type_names = MALLOC_N(char *, allocmax);
   scheme_type_readers = MALLOC_N_ATOMIC(Scheme_Type_Reader, allocmax);
@@ -275,6 +275,8 @@ scheme_init_type (Scheme_Env *env)
 #ifdef MZ_GC_BACKTRACE
   set_name(scheme_rt_meta_cont, "<meta-continuation>");
 #endif
+  set_name(scheme_place_type, "<place>");
+  set_name(scheme_engine_type, "<engine>");
 }
 
 Scheme_Type scheme_make_type(const char *name)
@@ -575,7 +577,7 @@ void scheme_register_traversers(void)
   GC_REG_TRAV(scheme_svector_type, svector_val);
 
   GC_REG_TRAV(scheme_set_macro_type, small_object);
-  GC_REG_TRAV(scheme_id_macro_type, small_object);
+  GC_REG_TRAV(scheme_id_macro_type, twoptr_obj);
 
   GC_REG_TRAV(scheme_stx_type, stx_val);
   GC_REG_TRAV(scheme_stx_offset_type, stx_off_val);
@@ -617,6 +619,9 @@ void scheme_register_traversers(void)
   GC_REG_TRAV(scheme_log_reader_type, mark_log_reader);
 
   GC_REG_TRAV(scheme_rt_runstack, runstack_val);
+
+  GC_REG_TRAV(scheme_rib_delimiter_type, small_object);
+  GC_REG_TRAV(scheme_noninline_proc_type, small_object);
 }
 
 END_XFORM_SKIP;
